import json
from collections import Iterable
from typing import Union

import yaml

DELIM = ","


class Config:
    _identify = None
    _overwritten = {}

    @staticmethod
    def handle_iterables(x):
        if not x:
            return x
        # if only one
        if isinstance(x, str):
            return [x]
        return x

    def __init__(self, **kwargs):
        for f, v in self:
            if f in kwargs:
                kv = kwargs.get(f)
                if v != kv:
                    setattr(self, f, kv)
                    self._overwritten[f] = v

    def __iter__(self):
        for k in set(self.__class__.__dict__.keys()).union(set(self.__dict__.keys())):
            if k[0] != "_" and (
                hasattr(getattr(self, k), "_identify") or not callable(getattr(self, k))
            ):
                yield k, getattr(self, k)

    def __str__(self):
        # or, i could have simply done a yaml dumps
        string = ""
        for k, v in self:
            print(k, v)
        for k, v in self:
            if hasattr(v, "_identify"):
                string += f"{k}:\n"
                string += "".join([f"--{vsub}\n" for vsub in str(v).split("\n")])
            else:
                string += f"{k}:{v}\n"
        return string[:-1]

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def parse(arg):
        if isinstance(arg, str) and DELIM in arg:
            arg = arg.split(DELIM)
            if len(arg) == 0:
                arg = ""
            else:
                arg = tuple(arg)
        elif isinstance(arg, str) and arg.isdigit():
            return int(arg)
        elif isinstance(arg, str) and arg.lower() == "true":
            arg = True
        elif isinstance(arg, str) and arg.lower() == "false":
            arg = False
        return arg

    def to_dict(self):
        data = {}
        for k, v in self:
            if hasattr(v, "_identify"):
                data[k] = v.to_dict()
            else:
                data[k] = v
        return data

    def dump_json(self, file):
        json.dump(self.to_dict(), open(file, "w"))

    def dump_yaml(self, file):
        yaml.dump(self.to_dict(), open(file, "w"), default_flow_style=False)

    @classmethod
    def load(cls, fp_name_dict: Union[str, dict]):
        raise NotImplementedError()

    @classmethod
    def from_dict(cls, config_dict):
        config = cls()
        config.update(config_dict)
        return config

    def update(self, updates: dict):
        for k, orig_v in self:
            if k in updates:
                v = updates.get(k)
                # I hardcode logdir and testrun as needed in all subconfigs
                if (k == "logdir" or k == "test_run") and hasattr(orig_v, "_identify"):
                    try:
                        setattr(orig_v, k, v)
                    except Exception:
                        raise Exception(orig_v, k, v)
                if isinstance(v, dict) and hasattr(orig_v, "_identify"):
                    orig_v.update(v)
                else:
                    if not isinstance(orig_v, str) and isinstance(orig_v, Iterable):
                        if isinstance(v, Iterable):
                            setattr(self, k, orig_v + v)
                        else:
                            setattr(self, k, orig_v.add(v))
                    else:
                        setattr(self, k, v)

    def list_subconfigs(self):
        subconfig_names = []
        for k, orig_v in self:
            if hasattr(orig_v, "_identify"):
                subconfig_names.append(k)
        return subconfig_names



import inspect
# note if we do not immport a pacakage correctly in this class, no loops or exps will be present
import json
import os
import resource
import sys
from collections import defaultdict
from itertools import chain

import torch
import vltk.vars as vltk
from datasets.utils.logging import set_verbosity_error
# disable logging from datasets
from vltk.dataset.basedataset import BaseDataset, CollatedVisionSets
from vltk.processing import Processors, VisnProcessor
from vltk.utils import base
from vltk.utils.adapters import (get_rawsize, get_scale, get_size,
                                 imagepoints_to_mask, rescale_box,
                                 resize_binary_mask, seg_to_mask)

__import__("tokenizers")
TOKENIZERS = {
    m[0]: m[1] for m in inspect.getmembers(sys.modules["tokenizers"], inspect.isclass)
}

rlimit = resource.getrlimit(resource.RLIMIT_NOFILE)
resource.setrlimit(resource.RLIMIT_NOFILE, (6144, rlimit[1]))

set_verbosity_error()


os.environ["TOKENIZERS_PARALLELISM"] = "False"


class VisionDataset(BaseDataset):
    _supported = (
        vltk.text,
        vltk.polygons,
        vltk.size,
        vltk.area,
        vltk.boxes,
        vltk.RLE,
    )

    def __init__(
        self,
        config,
        visndatasetadapterdict,
        annotationdict=None,
        metadata_ids=None,
        is_train=False,
        batch_info=None,
        tokenizer_in_visn_dataset=False,
        **kwargs,
    ):

        self.batch_info = batch_info
        self.tokenizer_in_visn_dataset = tokenizer_in_visn_dataset
        if tokenizer_in_visn_dataset:
            self._init_tokenizer(config.lang)
        else:
            self.tokenizer = None
            self.from_transformers = None
        self.is_train = is_train
        # self.annotationdict = annotationdict
        self.config = config
        self.metadata_ids = metadata_ids
        self.annotationdict = annotationdict
        self.metadata_counts = defaultdict(dict)
        for name in annotationdict:
            for meta_name, meta_counts in annotationdict[name]._meta_dict.items():
                self.metadata_counts[name][meta_name] = meta_counts

        self._init_image_processor(config)
        self._init_vision_processors(config)
        self.visndatasetadapterdict = visndatasetadapterdict
        # later if we need
        # print(len(set(annotationdict["funsdaug"].imgids)))

        annotationdict, imgids2pathes, n_imgs = self._shrink_annotation_dicts(
            annotationdict, visndatasetadapterdict
        )
        # print(len(set(annotationdict["funsdaug"].imgids)))

        self._init_annotation_dict(config, annotationdict)
        self.img_id_to_path = imgids2pathes
        self.n_imgs = n_imgs

    def _shrink_annotation_dicts(self, annotationdict, visndatasetadapterdict):
        imgids2pathes = {}
        n_imgs = 0
        for name, imgsetsplits in visndatasetadapterdict.items():
            annodata = annotationdict[name]
            imgids = set()
            for imgids2files in imgsetsplits.values():
                imgids.update(set(imgids2files.keys()))
                imgids2pathes.update(imgids2files)
            filtered_annos = annodata.imgid_filter(imgids, True)
            n_imgs += len(filtered_annos)
            # print("n_imgs", n_imgs)
            annotationdict[name] = filtered_annos
        return annotationdict, imgids2pathes, n_imgs

    @property
    def image(self):
        return self._image

    @property
    def annotations(self):
        return self._annotations

    def _init_vision_processors(self, config):
        vision_processors = (
            config.visn_processors if config.visn_processors is not None else []
        )
        vision_processors = [
            x if not isinstance(x, str) else Processors().get(x)
            for x in vision_processors
        ]

        vision_processors = list(
            filter(lambda x: x.__bases__[0] == VisnProcessor, vision_processors)
        )

        self.vision_processors = [
            x(
                tokenizer=self.tokenizer,
                from_transformers=self.from_transformers,
                config=self.config,
                metadata_ids=self.metadata_ids,
            )
            for x in vision_processors
        ]

        self.vision_processor_keys = ()
        for x in self.vision_processors:
            self.vision_processor_keys += tuple(x.keys)

    def run_vision_processors(self, entry):
        for processor in self.vision_processors:
            entry = processor(entry, config=self.config)
        return entry

    def _init_annotation_dict(self, config, annotationdict):
        if annotationdict is None or config.ignore_annotations:
            self._annotations = None
        else:
            annotations = list(annotationdict.values())
            self._annotations = CollatedVisionSets(*annotations)

    def _init_image_processor(self, config):
        if config.extractor is None:
            processor = config.visn.build()
            self._image = processor
            self._transforms = self._image.transforms

    @property
    def transforms(self):
        return {t.__class__.__name__: t for t in self._transforms}

    def _handle_image(self, entry):
        img_id = entry[vltk.imgid]
        if vltk.filepath not in entry:
            try:
                filepath = self.img_id_to_path[img_id]
            except KeyError:
                raise Exception(entry)
            entry[vltk.filepath] = filepath
        else:
            filepath = entry[vltk.filepath]
        if self.config.rand_feats is not None:
            feat_shape = tuple(self.config.rand_feats)
            img = torch.rand(feat_shape)
            entry[vltk.img] = img
        else:
            if not self.config.ignore_filepath:
                entry[vltk.filepath] = filepath
            else:
                entry.pop(vltk.filepath)
            entry[vltk.img] = self.image(filepath)

        entry[vltk.size] = get_size(self.image)
        entry[vltk.rawsize] = get_rawsize(self.image)
        if torch.all(entry[vltk.size].eq(entry[vltk.rawsize])):
            entry.pop(vltk.rawsize)
        entry[vltk.scale] = get_scale(self.image)

        if self.config.ignore_image:
            entry.pop(vltk.img)

        return entry

    @torch.no_grad()
    def _handle_annotations(self, entry, replace_keys=None):
        skip_segmentation = (
            True
            if (vltk.size not in entry or self.config.ignore_segmentation)
            else False
        )
        # get annotations for image
        # entry.update(self.annotations.get(img_id))

        if skip_segmentation and vltk.polygons in entry:
            entry.pop(vltk.polygons)
        if skip_segmentation and vltk.RLE in entry:
            entry.pop(vltk.RLE)

        # run vision processors
        entry = self.run_vision_processors(entry)

        if replace_keys is not None:
            for r in replace_keys:
                if r in entry:
                    entry[vltk.VLOVERLAP[r]] = entry[r]

        return entry

    def __len__(self):
        return self.n_imgs

    @torch.no_grad()
    def __getitem__(self, i):
        anno_dict, anno_dataset = self.annotations[i]
        anno_dict = self._handle_image(anno_dict)
        if self.annotations is not None:
            anno_dict = self._handle_annotations(anno_dict)
        anno_dict = self.try_tensorify(anno_dict)
        self.batch_info.update_entry_keys(anno_dict)
        return anno_dict



from collections import defaultdict

import vltk.vars as vltk
from tqdm import tqdm
from vltk import adapters
from vltk.adapters import Adapters
from vltk.configs import DataConfig
from vltk.features import Features


# data source: https://github.com/ccvl/clevr-refplus-dataset-gen
# https://cs.jhu.edu/~cxliu/data/clevr_ref+_1.0.zip
class CLEVRREF(adapters.VisnDataset):
    @staticmethod
    def schema():
        return {
            vltk.RLE: Features.RLE(),
            "colors": Features.StringList(),
            "shapes": Features.StringList(),
            "sizes": Features.StringList(),
            "materials": Features.StringList(),
            vltk.box: Features.Box(),
        }

    @staticmethod
    def forward(json_files, splits):
        # default box order: x, y, h, w
        entries = defaultdict(dict)
        for filepath, js in json_files.items():
            if "scene" not in filepath:
                continue
            for i, scene in enumerate(tqdm(js["scenes"])):
                img_filename = scene["image_filename"]
                imgid = img_filename.split(".")[0]
                colors = []
                shapes = []
                materials = []
                sizes = []
                boxes = []
                points = []
                for idx, (obj, bbox, seg) in enumerate(
                    zip(
                        scene["objects"],
                        scene["obj_bbox"].values(),
                        scene["obj_mask"].values(),
                    )
                ):
                    boxes.append(bbox)
                    colors.append(obj["color"])
                    shapes.append(obj["shape"])
                    materials.append(obj["material"])
                    sizes.append(obj["size"])
                    try:

                        seg = seg.split(",")
                    except Exception:
                        continue

                    points.append(seg)

                entries[imgid] = {
                    vltk.RLE: points,
                    "colors": colors,
                    "shapes": shapes,
                    "materials": materials,
                    "sizes": sizes,
                    vltk.box: boxes,
                    vltk.imgid: imgid,
                }

        return [v for v in entries.values()]


if __name__ == "__main__":
    # set datadir
    datadir = "/home/eltoto/demodata"
    # create datasets
    # clevrref = CLEVRREF.extract(datadir, ignore_files="exp")
    print(Adapters().avail())

    # TODO: make sure visnlang adapters are being added
    Adapters().add(CLEVRREF)
    # define config for dataset
    config = DataConfig(
        # choose which dataset and dataset split for train and eval
        train_datasets=[
            ["clevrref", "trainval"],
        ],
        # eval_datasets=["gqa", "testdev"],
        # choose which tokenizer to use
        tokenizer="BertWordPieceTokenizer",
        # choose which feature extractor to use
        extractor=None,
        datadir=datadir,
        train_batch_size=1,
        eval_batch_size=1,
        img_first=True,
    )



import json
import os
import pickle
from abc import abstractmethod
from collections import Counter, OrderedDict, defaultdict
from pathlib import Path

import datasets as ds
import pyarrow
import vltk.vars as vltk
from datasets import ArrowWriter
from tqdm import tqdm
from vltk.abc.adapter import Adapter
from vltk.inspection import collect_args_to_func
from vltk.utils.base import set_metadata, try_load


class VisnDataset(Adapter):
    _batch_size = 1024
    _base_features = {
        vltk.imgid: ds.Value("string"),
    }
    _meta_names = {"img_to_row_map", "object_frequencies", "vocab"}
    _is_annotation = True
    _extensions = {"jpg", "png", "jpeg"}

    @staticmethod
    def adjust_imgid(img_id, dataset_name=None):
        """
        Sometimes the image IDS provided in Vision datasets are repeated in different datasets. If that is the case, implementing
        this optional function will adjust the image id in aims to appropriately mactch
        the corresponging dataset try prepending the dataset name
        """
        return img_id

    @classmethod
    def filepath(cls, imgid, datadir, split):
        filename = cls.imgid_to_filename(imgid, split)
        return os.path.join(datadir, cls.name, split, filename)

    @classmethod
    def load_imgid2path(cls, datadir, split):
        name = cls.__name__.lower()
        return VisnDataset.files(datadir, name, split)

    @staticmethod
    def files(path, name, split):
        files = {}
        path = os.path.join(path, name)
        if not os.path.isdir(path):
            print(f"No path exists for: {path}")
            return files
        for g_ext in VisnDataset._extensions:
            for i in Path(path).glob(f"**/*.{g_ext}"):
                if i.is_dir():
                    continue
                stem = i.stem
                fp = str(i)
                if split == "":
                    cont = False
                    for spl in vltk.SPLITALIASES:
                        if spl in stem:
                            cont = True
                            break
                    if cont:
                        continue
                elif split not in fp:
                    continue

                # TODO: confirm if I still want to prepend dataset name later
                # okay, so we will only add the dataset name if it is not already present
                # actually, lets not worry about this until we run into this issue
                # if name.casefold() not in iid.casefold():
                #     iid = f'{name}{vltk.delim}{i.split(".")[0]}'
                # iid = i.split(".")[0]
                files[stem] = fp
        return files

    @classmethod
    def extract(
        cls,
        searchdir,
        savedir=None,
        data_format="jpg",
        ignore_files=None,
        **kwargs,
    ):

        schema_dict = collect_args_to_func(cls.schema, kwargs=kwargs)
        feature_dict = {**cls.schema(**schema_dict), **cls._base_features}
        # gather info from schema to figure out what metadata to collect
        meta_dict = cls._init_metadata(feature_dict)
        # lets work on doing the annotations first
        total_annos = {}
        searchdir, _ = cls._get_valid_search_pathes(
            searchdir, name=cls.__name__.lower(), annodir=vltk.ANNOTATION_DIR
        )
        files = cls._iter_files(searchdir)
        json_files = {}
        temp_splits = []
        print("loading annotations...")
        for anno_file in tqdm(files):
            if ignore_files is not None and ignore_files in str(anno_file):
                continue

            split = None
            for spl in vltk.SPLITALIASES:
                if spl in str(anno_file):
                    split = spl
                    break
            temp_splits.append(split)
            anno_data = try_load(anno_file)
            json_files[str(anno_file).split("/")[-1]] = anno_data

        kwargs["datadir"] = "/".join(searchdir.split("/")[:-2])
        forward_dict = collect_args_to_func(cls.forward, kwargs=kwargs)
        total_annos = cls.forward(json_files, temp_splits, **forward_dict)

        # now write
        print("writing to Datasets/Arrow object")
        writer, buffer, extra_meta = cls._write_batches(
            total_annos,
            feature_dict,
            cls._batch_size,
            cls.__name__.lower(),
            meta_dict=meta_dict,
        )
        if savedir is None:
            savedir = searchdir

        (table, meta_dict) = cls._write_data(writer, buffer, savedir, extra_meta)
        if table is None:
            return None
        return cls(arrow_table=table, meta_dict=meta_dict)

    @staticmethod
    def _write_batches(annos, feature_dict, batch_size, name, meta_dict):
        # name refers to the dataset (class) name
        # object_dict = Counter()
        features = ds.Features(feature_dict)
        imgid2rows = defaultdict(list)  # OrderedDict()
        extra_vocab = set()
        cur_size = 0
        cur_row = 0
        buffer = pyarrow.BufferOutputStream()
        stream = pyarrow.output_stream(buffer)
        writer = ArrowWriter(features=features, stream=stream)
        n_files = len(annos)
        # change feature types to classes isntead
        for i, entry in enumerate(annos):
            imgs_left = abs(i + 1 - n_files)

            entry[vltk.imgid] = VisnDataset.adjust_imgid(
                entry[vltk.imgid],
                name,
            )
            img_id = entry[vltk.imgid]
            if vltk.text in entry:
                extra_vocab.update(entry[vltk.text])
            meta_dict = VisnDataset._update_metadata(meta_dict, entry)
            imgid2rows[img_id].append(cur_row)
            cur_row += 1
            if cur_size == 0:
                for k, v in entry.items():
                    entry[k] = [v]
                cur_batch = entry
                cur_size = 1
            else:

                for k, v in entry.items():
                    cur_batch[k].append(v)
                cur_size += 1

            # write features
            if cur_size == batch_size or imgs_left < batch_size:
                cur_size = 0
                batch = features.encode_batch(cur_batch)
                writer.write_batch(batch)

        meta_dict["img_to_row_map"] = imgid2rows
        meta_dict["vocab"] = extra_vocab
        return writer, buffer, meta_dict

    @property
    def labels(self):
        return set(self._object_frequencies.keys())

    @staticmethod
    def _write_data(writer, buffer, savedir, extra_meta):
        print("saving...")
        value = buffer.getvalue()
        if value.size == 0:
            print("WARNING: no data saved")
            return (None, None)
            # do something
        dset = ds.Dataset.from_buffer(value)
        try:
            writer.finalize(close_stream=False)
        except Exception:
            pass
        # misc.
        dset = pickle.loads(pickle.dumps(dset))
        savefile = os.path.join(savedir, "annotations.arrow")

        # add extra metadata
        table = set_metadata(dset._data, tbl_meta=extra_meta)
        # define new writer
        writer = ArrowWriter(path=savefile, schema=table.schema, with_metadata=False)
        # savedir new table
        writer.write_table(table)
        e, b = VisnDataset._custom_finalize(writer, close_stream=True)
        print(f"Success! You wrote {e} entry(s) and {b >> 20} mb")
        print(f"Located: {savefile}")
        return (table, extra_meta)

    def align_imgids(self):
        for i in range(len(self)):
            self._img_to_row_map[self[i][vltk.imgid]] = i

    def check_imgid_alignment(self):
        orig_map = self.img_to_row_map
        for i in range(len(self)):
            img_id = self[i][vltk.imgid]
            mapped_ind = orig_map[img_id]
            if mapped_ind != i:
                return False
            self._img_to_row_map[self[i][vltk.imgid]] = i
        return True

    @abstractmethod
    def forward(json_files, **kwargs):
        raise Exception("child forward is not being called")

    @abstractmethod
    def schema(*args, **kwargs):
        return dict

    @property
    def vocab(self):
        if hasattr(self, "_vocab"):
            return set(self._vocab.decode().split("\n"))
        else:
            return None

    # @abstractmethod
    # def imgid_to_filename(imgid, split):
    #     return str



"""
 coding=utf-8
 Copyright 2018, Antonio Mendoza Hao Tan, Mohit Bansal
 Adapted From Facebook Inc, Detectron2 && Huggingface Co.

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.import copy
 """
import itertools
import math
import os
from abc import ABCMeta, abstractmethod
from collections import OrderedDict, namedtuple
from typing import Dict, List, Tuple

import numpy as np
import torch
from torch import nn
from torch.nn import functional as F
from torch.nn.modules.batchnorm import BatchNorm2d
from torchvision.ops import RoIPool
from torchvision.ops.boxes import batched_nms, nms
from vltk import decorators
from vltk.compat import (WEIGHTS_NAME, Config, cached_path, hf_bucket_url,
                         is_remote_url, load_checkpoint)

__all__ = ["FRCNN"]


# other:
# def norm_box(boxes, raw_sizes):
#     raise Exception(boxes)
#     if not isinstance(boxes, torch.Tensor):
#         normalized_boxes = boxes.copy()
#     else:
#         normalized_boxes = boxes.clone()
#     backtolist = False
#     if isinstance(normalized_boxes, list):
#         normalized_boxes = np.array(normalized_boxes)
#         backtolist = True
#         raise Exception(normalized_boxes)
#     normalized_boxes[:, 1::2] /= raw_sizes[:, 0].cpu().view(-1,1,1)
#     normalized_boxes[:, 0::2] /= raw_sizes[:, 1].cpu().view(-1,1,1)
#     if backtolist:
#         normalized_boxes = list(normalized_boxes)

#     return normalized_boxes


# def pad_list_tensors(list_tensors, preds_per_image, max_detections=None,
#         return_tensors=None, padding=None, pad_value=0, location = None):
#     """
#     location will always be cpu for np tensors
#     """
#     if location is None:
#         location = "cpu"
#     assert return_tensors in {"pt", "np", None}
#     assert padding in {"max_detections", "max_batch", None}
#     new = []
#     if padding is None:
#         if return_tensors is None:
#             return list_tensors
#         elif return_tensors == "pt":
#             if not isinstance(list_tensors, torch.Tensor):
#                 return torch.stack(list_tensors).to(location)
#             else:
#                 return list_tensors.to(location)
#         else:
#             if not isinstance(list_tensors, list):
#                return np.array(list_tensors.to(location))
#             else:
#                return list_tensors.to(location)
#     if padding == "max_detections":
#         assert max_detections is not None, "specify max number of detections per batch"
#     elif padding == "max_batch":
#         max_detections = max(preds_per_image)
#     for i in range(len(list_tensors)):
#         too_small = False
#         tensor_i = list_tensors.pop(0)
#         if tensor_i.ndim < 2:
#             too_small = True
#             tensor_i = tensor_i.unsqueeze(-1)
#         assert isinstance(tensor_i, torch.Tensor)
#         tensor_i = F.pad(input=tensor_i, pad=(0, 0, 0, max_detections - preds_per_image[i]), mode="constant", value=pad_value)
#         if too_small:
#             tensor_i = tensor_i.squeeze(-1)
#         if return_tensors is None:
#             if location == "cpu":
#                 tensor_i = tensor_i.cpu()
#             tensor_i = tensor_i.tolist()
#         if return_tensors == "np":
#             if location == "cpu":
#                 tensor_i = tensor_i.cpu()
#             tensor_i = tensor_i.numpy()
#         else:
#             if location == "cpu" and isinstance(tensor_i, torch.Tensor):
#                 tensor_i = tensor_i.cpu()
#         new.append(tensor_i)
#     if return_tensors == "np":
#         return np.stack(new, axis=0)
#     elif return_tensors == "pt" and not isinstance(new, torch.Tensor):
#         return torch.stack(new, dim=0)
#     else:
#         return list_tensors


def do_nms(boxes, scores, image_shape, score_thresh, nms_thresh, mind, maxd):
    scores = scores[:, :-1]
    num_bbox_reg_classes = boxes.shape[1] // 4
    # Convert to Boxes to use the `clip` function ...
    boxes = boxes.reshape(-1, 4)
    _clip_box(boxes, image_shape)
    boxes = boxes.view(-1, num_bbox_reg_classes, 4)  # R x C x 4

    # Select max scores
    max_scores, max_classes = scores.max(1)       # R x C --> R
    num_objs = boxes.size(0)
    boxes = boxes.view(-1, 4)
    idxs = torch.arange(num_objs).to(boxes.device) * num_bbox_reg_classes + max_classes
    max_boxes = boxes[idxs]     # Select max boxes according to the max scores.

    # Apply NMS
    keep = nms(max_boxes, max_scores, nms_thresh)
    keep = keep[:maxd]
    if keep.shape[-1] >= mind and keep.shape[-1] <= maxd:
        stop = True
        max_boxes, max_scores = max_boxes[keep], max_scores[keep]
        classes = max_classes[keep]
        return stop, max_boxes, max_scores, classes, keep
    else:
        stop = False
        max_boxes, max_scores = max_boxes[keep], max_scores[keep]
        classes = max_classes[keep]
        return stop, max_boxes, max_scores, classes, keep


# Helper Functions
def _clip_box(tensor, box_size: Tuple[int, int]):
    assert torch.isfinite(tensor).all(), "Box tensor contains infinite or NaN!"
    h, w = box_size
    tensor[:, 0].clamp_(min=0, max=w)
    tensor[:, 1].clamp_(min=0, max=h)
    tensor[:, 2].clamp_(min=0, max=w)
    tensor[:, 3].clamp_(min=0, max=h)


def _nonempty_boxes(box, threshold: float = 0.0) -> torch.Tensor:
    widths = box[:, 2] - box[:, 0]
    heights = box[:, 3] - box[:, 1]
    keep = (widths > threshold) & (heights > threshold)
    return keep


def get_norm(norm, out_channels):
    if isinstance(norm, str):
        if len(norm) == 0:
            return None
        norm = {
            "BN": BatchNorm2d,
            "GN": lambda channels: nn.GroupNorm(32, channels),
            "nnSyncBN": nn.SyncBatchNorm,  # keep for debugging
            "": lambda x: x,
        }[norm]
    return norm(out_channels)


def _create_grid_offsets(size: List[int], stride: int, offset: float, device):

    grid_height, grid_width = size
    shifts_x = torch.arange(
        offset * stride,
        grid_width * stride,
        step=stride,
        dtype=torch.float32,
        device=device,
    )
    shifts_y = torch.arange(
        offset * stride,
        grid_height * stride,
        step=stride,
        dtype=torch.float32,
        device=device,
    )

    shift_y, shift_x = torch.meshgrid(shifts_y, shifts_x)
    shift_x = shift_x.reshape(-1)
    shift_y = shift_y.reshape(-1)
    return shift_x, shift_y


def build_backbone(cfg):
    input_shape = ShapeSpec(channels=len(cfg.MODEL.PIXEL_MEAN))
    norm = cfg.RESNETS.NORM
    stem = BasicStem(
        in_channels=input_shape.channels,
        out_channels=cfg.RESNETS.STEM_OUT_CHANNELS,
        norm=norm,
        caffe_maxpool=cfg.MODEL.MAX_POOL,
    )
    freeze_at = cfg.BACKBONE.FREEZE_AT

    if freeze_at >= 1:
        for p in stem.parameters():
            p.requires_grad = False

    out_features = cfg.RESNETS.OUT_FEATURES
    depth = cfg.RESNETS.DEPTH
    num_groups = cfg.RESNETS.NUM_GROUPS
    width_per_group = cfg.RESNETS.WIDTH_PER_GROUP
    bottleneck_channels = num_groups * width_per_group
    in_channels = cfg.RESNETS.STEM_OUT_CHANNELS
    out_channels = cfg.RESNETS.RES2_OUT_CHANNELS
    stride_in_1x1 = cfg.RESNETS.STRIDE_IN_1X1
    res5_dilation = cfg.RESNETS.RES5_DILATION
    assert res5_dilation in {1, 2}, "res5_dilation cannot be {}.".format(res5_dilation)

    num_blocks_per_stage = {50: [3, 4, 6, 3], 101: [3, 4, 23, 3], 152: [3, 8, 36, 3]}[
        depth
    ]

    stages = []
    out_stage_idx = [
        {"res2": 2, "res3": 3, "res4": 4, "res5": 5}[f] for f in out_features
    ]
    max_stage_idx = max(out_stage_idx)
    for idx, stage_idx in enumerate(range(2, max_stage_idx + 1)):
        dilation = res5_dilation if stage_idx == 5 else 1
        first_stride = 1 if idx == 0 or (stage_idx == 5 and dilation == 2) else 2
        stage_kargs = {
            "num_blocks": num_blocks_per_stage[idx],
            "first_stride": first_stride,
            "in_channels": in_channels,
            "bottleneck_channels": bottleneck_channels,
            "out_channels": out_channels,
            "num_groups": num_groups,
            "norm": norm,
            "stride_in_1x1": stride_in_1x1,
            "dilation": dilation,
        }

        stage_kargs["block_class"] = BottleneckBlock
        blocks = ResNet.make_stage(**stage_kargs)
        in_channels = out_channels
        out_channels *= 2
        bottleneck_channels *= 2

        if freeze_at >= stage_idx:
            for block in blocks:
                block.freeze()
        stages.append(blocks)

    return ResNet(stem, stages, out_features=out_features)


def find_top_rpn_proposals(
    proposals,
    pred_objectness_logits,
    images,
    image_sizes,
    nms_thresh,
    pre_nms_topk,
    post_nms_topk,
    min_box_side_len=15,
    training=False,
    scales_yx=None,
    ignorey=None,
):
    """Args:
        proposals (list[Tensor]): (L, N, Hi*Wi*A, 4).
        pred_objectness_logits: tensors of lenngth L.
        nms_thresh (float): IoU threshold to use for NMS
        pre_nms_topk (int): before nms
        post_nms_topk (int): after nms
        min_box_side_len (float): minimum proposal box side
        training (bool): True if proposals are to be used in training,
    Returns:
        resuls (List[Dict]): stores post_nms_topk object proposals for image i.
    """
    num_images = len(images)
    device = proposals[0].device

    # 1. Select top-k anchor for every level and every image
    topk_scores = []  # #lvl Tensor, each of shape N x topk
    topk_proposals = []
    level_ids = []  # #lvl Tensor, each of shape (topk,)
    batch_idx = torch.arange(num_images, device=device)
    for level_id, proposals_i, logits_i in zip(
        itertools.count(), proposals, pred_objectness_logits
    ):
        Hi_Wi_A = logits_i.shape[1]
        num_proposals_i = min(pre_nms_topk, Hi_Wi_A)

        # sort is faster than topk (https://github.com/pytorch/pytorch/issues/22812)
        # topk_scores_i, topk_idx = logits_i.topk(num_proposals_i, dim=1)
        logits_i, idx = logits_i.sort(descending=True, dim=1)
        topk_scores_i = logits_i[batch_idx, :num_proposals_i]
        topk_idx = idx[batch_idx, :num_proposals_i]

        # each is N x topk
        topk_proposals_i = proposals_i[batch_idx[:, None], topk_idx]  # N x topk x 4
        topk_proposals.append(topk_proposals_i)
        topk_scores.append(topk_scores_i)
        level_ids.append(
            torch.full((num_proposals_i,), level_id, dtype=torch.int64, device=device)
        )

    # 2. Concat all levels together
    topk_scores = torch.cat(topk_scores, dim=1)
    topk_proposals = torch.cat(topk_proposals, dim=1)
    level_ids = torch.cat(level_ids, dim=0)

    # if I change to batched_nms, I wonder if this will make a difference
    # 3. For each image, run a per-level NMS, and choose topk results.
    results = []
    for n, image_size in enumerate(image_sizes):
        boxes = topk_proposals[n]
        scores_per_img = topk_scores[n]
        # ignore y-range here
        if ignorey is not None and scales_yx is not None:
            assert ignorey.ndim == 3
            for ignoreyij in ignorey[n]:
                ignoreyij = ignoreyij.unsqueeze(0).repeat_interleave(boxes.size(0), dim=0) * 1/(scales_yx[n, 1])

                keep = torch.bitwise_and(torch.le(ignoreyij[:,1], boxes[:,3]),
                    torch.ge(ignoreyij[:,0], boxes[:,1])
                )
                keep= keep== 0
                boxes = boxes[keep]
                ignoreyij = ignoreyij[keep]
                scores_per_img= scores_per_img[keep]
                level_ids = level_ids[keep]

                box_ignore_above =  torch.bitwise_and(torch.gt(boxes[:,1], ignoreyij[:,
                    1]), torch.gt(boxes[:,3], ignoreyij[:,0]))
                box_ignore_below =  torch.bitwise_and(
                    torch.le(boxes[:,3], ignoreyij[:, 0]),
                    torch.gt(boxes[:,3], ignoreyij[:,0])
                )
                ignore_clip_range= torch.bitwise_or(box_ignore_below,box_ignore_above)
                to_clip = ignore_clip_range == 0

                clip_top= torch.bitwise_and(
                    to_clip,
                    torch.lt(
                        torch.abs(ignoreyij[:,1] - boxes[:,3]),
                        torch.abs(ignoreyij[:,0] - boxes[:, 1]),
                    )
                )
                clip_bottom= torch.bitwise_and(
                    to_clip,
                    torch.lt(
                        torch.abs(ignoreyij[:,0] - boxes[:, 1]),
                        torch.abs(ignoreyij[:,1] - boxes[:,3]),
                    )
                )
                boxes[clip_bottom, 1] = int(ignoreyij[0, 1])
                boxes[clip_top, 3] = int(ignoreyij[0, 0])

        # I will have to take a look at the boxes clip method
        _clip_box(boxes, image_size)
        # filter empty boxes
        keep = _nonempty_boxes(boxes, threshold=min_box_side_len)


        lvl = level_ids
        if keep.sum().item() != len(boxes):
            boxes, scores_per_img, lvl = (
                boxes[keep],
                scores_per_img[keep],
                level_ids[keep],
            )


        keep = batched_nms(boxes, scores_per_img, lvl, nms_thresh)
        keep = keep[:post_nms_topk]

        res = (boxes[keep], scores_per_img[keep])
        results.append(res)

    # I wonder if it would be possible for me to pad all these things.
    return results


def subsample_labels(labels, num_samples, positive_fraction, bg_label):
    """
    Returns:
        pos_idx, neg_idx (Tensor):
            1D vector of indices. The total length of both is `num_samples` or fewer.
    """
    positive = torch.nonzero((labels != -1) & (labels != bg_label)).squeeze(1)
    negative = torch.nonzero(labels == bg_label).squeeze(1)

    num_pos = int(num_samples * positive_fraction)
    # protect against not enough positive examples
    num_pos = min(positive.numel(), num_pos)
    num_neg = num_samples - num_pos
    # protect against not enough negative examples
    num_neg = min(negative.numel(), num_neg)

    # randomly select positive and negative examples
    perm1 = torch.randperm(positive.numel(), device=positive.device)[:num_pos]
    perm2 = torch.randperm(negative.numel(), device=negative.device)[:num_neg]

    pos_idx = positive[perm1]
    neg_idx = negative[perm2]
    return pos_idx, neg_idx


def add_ground_truth_to_proposals(gt_boxes, proposals):
    raise NotImplementedError()


def add_ground_truth_to_proposals_single_image(gt_boxes, proposals):
    raise NotImplementedError()


def _fmt_box_list(box_tensor, batch_index: int):
    repeated_index = torch.full(
        (len(box_tensor), 1),
        batch_index,
        dtype=box_tensor.dtype,
        device=box_tensor.device,
    )
    return torch.cat((repeated_index, box_tensor), dim=1)


def convert_boxes_to_pooler_format(box_lists: List[torch.Tensor]):
    pooler_fmt_boxes = torch.cat(
        [_fmt_box_list(box_list, i) for i, box_list in enumerate(box_lists)],
        dim=0,
    )
    return pooler_fmt_boxes


def assign_boxes_to_levels(
    box_lists: List[torch.Tensor],
    min_level: int,
    max_level: int,
    canonical_box_size: int,
    canonical_level: int,
):

    box_sizes = torch.sqrt(torch.cat([boxes.area() for boxes in box_lists]))
    # Eqn.(1) in FPN paper
    level_assignments = torch.floor(
        canonical_level + torch.log2(box_sizes / canonical_box_size + 1e-8)
    )
    # clamp level to (min, max), in case the box size is too large or too small
    # for the available feature maps
    level_assignments = torch.clamp(level_assignments, min=min_level, max=max_level)
    return level_assignments.to(torch.int64) - min_level


# Helper Classes
class _NewEmptyTensorOp(torch.autograd.Function):
    @ staticmethod
    def forward(ctx, x, new_shape):
        ctx.shape = x.shape
        return x.new_empty(new_shape)

    @ staticmethod
    def backward(ctx, grad):
        shape = ctx.shape
        return _NewEmptyTensorOp.apply(grad, shape), None


class ShapeSpec(namedtuple("_ShapeSpec", ["channels", "height", "width", "stride"])):
    def __new__(cls, *, channels=None, height=None, width=None, stride=None):
        return super().__new__(cls, channels, height, width, stride)


class Box2BoxTransform(object):
    """
    This R-CNN transformation scales the box's width and height
    by exp(dw), exp(dh) and shifts a box's center by the offset
    (dx * width, dy * height).
    """

    def __init__(
        self, weights: Tuple[float, float, float, float], scale_clamp: float = None
    ):
        """
        Args:
            weights (4-element tuple): Scaling factors that are applied to the
                (dx, dy, dw, dh) deltas. In Fast R-CNN, these were originally set
                such that the deltas have unit variance; now they are treated as
                hyperparameters of the system.
            scale_clamp (float): When predicting deltas, the predicted box scaling
                factors (dw and dh) are clamped such that they are <= scale_clamp.
        """
        self.weights = weights
        if scale_clamp is not None:
            self.scale_clamp = scale_clamp
        else:
            """
            Value for clamping large dw and dh predictions.
            The heuristic is that we clamp such that dw and dh are no larger
            than what would transform a 16px box into a 1000px box
            (based on a small anchor, 16px, and a typical image size, 1000px).
            """
            self.scale_clamp = math.log(1000.0 / 16)

    def get_deltas(self, src_boxes, target_boxes):
        """
        Get box regression transformation deltas (dx, dy, dw, dh) that can be used
        to transform the `src_boxes` into the `target_boxes`. That is, the relation
        ``target_boxes == self.apply_deltas(deltas, src_boxes)`` is true (unless
        any delta is too large and is clamped).
        Args:
            src_boxes (Tensor): source boxes, e.g., object proposals
            target_boxes (Tensor): target of the transformation, e.g., ground-truth
                boxes.
        """
        assert isinstance(src_boxes, torch.Tensor), type(src_boxes)
        assert isinstance(target_boxes, torch.Tensor), type(target_boxes)

        src_widths = src_boxes[:, 2] - src_boxes[:, 0]
        src_heights = src_boxes[:, 3] - src_boxes[:, 1]
        src_ctr_x = src_boxes[:, 0] + 0.5 * src_widths
        src_ctr_y = src_boxes[:, 1] + 0.5 * src_heights

        target_widths = target_boxes[:, 2] - target_boxes[:, 0]
        target_heights = target_boxes[:, 3] - target_boxes[:, 1]
        target_ctr_x = target_boxes[:, 0] + 0.5 * target_widths
        target_ctr_y = target_boxes[:, 1] + 0.5 * target_heights

        wx, wy, ww, wh = self.weights
        dx = wx * (target_ctr_x - src_ctr_x) / src_widths
        dy = wy * (target_ctr_y - src_ctr_y) / src_heights
        dw = ww * torch.log(target_widths / src_widths)
        dh = wh * torch.log(target_heights / src_heights)

        deltas = torch.stack((dx, dy, dw, dh), dim=1)
        assert (
            (src_widths > 0).all().item()
        ), "Input boxes to Box2BoxTransform are not valid!"
        return deltas

    def apply_deltas(self, deltas, boxes, ignorey=None, scales_yx=None):
        """
        Apply transformation `deltas` (dx, dy, dw, dh) to `boxes`.
        Args:
            deltas (Tensor): transformation deltas of shape (N, k*4), where k >= 1.
                deltas[i] represents k potentially different class-specific
                box transformations for the single box boxes[i].
            boxes (Tensor): boxes to transform, of shape (N, 4)
        """
        boxes = boxes.to(deltas.dtype)

        widths = boxes[:, 2] - boxes[:, 0]
        heights = boxes[:, 3] - boxes[:, 1]
        ctr_x = boxes[:, 0] + 0.5 * widths
        ctr_y = boxes[:, 1] + 0.5 * heights

        wx, wy, ww, wh = self.weights
        dx = deltas[:, 0::4] / wx
        dy = deltas[:, 1::4] / wy
        dw = deltas[:, 2::4] / ww
        dh = deltas[:, 3::4] / wh

        # Prevent sending too large values into torch.exp()
        dw = torch.clamp(dw, max=self.scale_clamp)
        dh = torch.clamp(dh, max=self.scale_clamp)

        pred_ctr_x = dx * widths[:, None] + ctr_x[:, None]
        pred_ctr_y = dy * heights[:, None] + ctr_y[:, None]
        pred_w = torch.exp(dw) * widths[:, None]
        pred_h = torch.exp(dh) * heights[:, None]

        pred_boxes = torch.zeros_like(deltas)
        pred_boxes[:, 0::4] = pred_ctr_x - 0.5 * pred_w  # x1
        pred_boxes[:, 1::4] = pred_ctr_y - 0.5 * pred_h  # y1
        pred_boxes[:, 2::4] = pred_ctr_x + 0.5 * pred_w  # x2
        pred_boxes[:, 3::4] = pred_ctr_y + 0.5 * pred_h  # y2
        return pred_boxes


class Matcher(object):
    """
    This class assigns to each predicted "element" (e.g., a box) a ground-truth
    element. Each predicted element will have exactly zero or one matches; each
    ground-truth element may be matched to zero or more predicted elements.
    The matching is determined by the MxN match_quality_matrix, that characterizes
    how well each (ground-truth, prediction)-pair match each other. For example,
    if the elements are boxes, this matrix may contain box intersection-over-union
    overlap values.
    The matcher returns (a) a vector of length N containing the index of the
    ground-truth element m in [0, M) that matches to prediction n in [0, N).
    (b) a vector of length N containing the labels for each prediction.
    """

    def __init__(
        self,
        thresholds: List[float],
        labels: List[int],
        allow_low_quality_matches: bool = False,
    ):
        """
        Args:
            thresholds (list): a list of thresholds used to stratify predictions
                into levels.
            labels (list): a list of values to label predictions belonging at
                each level. A label can be one of {-1, 0, 1} signifying
                {ignore, negative class, positive class}, respectively.
            allow_low_quality_matches (bool): if True, produce additional matches or predictions with maximum match quality lower than high_threshold.
                For example, thresholds = [0.3, 0.5] labels = [0, -1, 1] All predictions with iou < 0.3 will be marked with 0 and
                thus will be considered as false positives while training. All predictions with 0.3 <= iou < 0.5 will be marked with -1 and
                thus will be ignored. All predictions with 0.5 <= iou will be marked with 1 and thus will be considered as true positives.
        """
        thresholds = thresholds[:]
        assert thresholds[0] > 0
        thresholds.insert(0, -float("inf"))
        thresholds.append(float("inf"))
        assert all(
            [low <= high for (low, high) in zip(thresholds[:-1], thresholds[1:])]
        )
        assert all([label_i in [-1, 0, 1] for label_i in labels])
        assert len(labels) == len(thresholds) - 1
        self.thresholds = thresholds
        self.labels = labels
        self.allow_low_quality_matches = allow_low_quality_matches

    def __call__(self, match_quality_matrix):
        """
        Args:
            match_quality_matrix (Tensor[float]): an MxN tensor, containing the pairwise quality between M ground-truth elements and N predicted
                elements. All elements must be >= 0 (due to the us of `torch.nonzero` for selecting indices in :meth:`set_low_quality_matches_`).
        Returns:
            matches (Tensor[int64]): a vector of length N, where matches[i] is a matched ground-truth index in [0, M)
            match_labels (Tensor[int8]): a vector of length N, where pred_labels[i] indicates true or false positive or ignored
        """
        assert match_quality_matrix.dim() == 2
        if match_quality_matrix.numel() == 0:
            default_matches = match_quality_matrix.new_full(
                (match_quality_matrix.size(1),), 0, dtype=torch.int64
            )
            # When no gt boxes exist, we define IOU = 0 and therefore set labels
            # to `self.labels[0]`, which usually defaults to background class 0
            # To choose to ignore instead,
            # can make labels=[-1,0,-1,1] + set appropriate thresholds
            default_match_labels = match_quality_matrix.new_full(
                (match_quality_matrix.size(1),), self.labels[0], dtype=torch.int8
            )
            return default_matches, default_match_labels

        assert torch.all(match_quality_matrix >= 0)

        # match_quality_matrix is M (gt) x N (predicted)
        # Max over gt elements (dim 0) to find best gt candidate for each prediction
        matched_vals, matches = match_quality_matrix.max(dim=0)

        match_labels = matches.new_full(matches.size(), 1, dtype=torch.int8)

        for (l, low, high) in zip(
            self.labels, self.thresholds[:-1], self.thresholds[1:]
        ):
            low_high = (matched_vals >= low) & (matched_vals < high)
            match_labels[low_high] = l

        if self.allow_low_quality_matches:
            self.set_low_quality_matches_(match_labels, match_quality_matrix)

        return matches, match_labels

    def set_low_quality_matches_(self, match_labels, match_quality_matrix):
        """
        Produce additional matches for predictions that have only low-quality matches.
        Specifically, for each ground-truth G find the set of predictions that have
        maximum overlap with it (including ties); for each prediction in that set, if
        it is unmatched, then match it to the ground-truth G.
        This function implements the RPN assignment case (i)
        in Sec. 3.1.2 of Faster R-CNN.
        """
        # For each gt, find the prediction with which it has highest quality
        highest_quality_foreach_gt, _ = match_quality_matrix.max(dim=1)
        # Find the highest quality match available, even if it is low, including ties.
        # Note that the matches qualities must be positive due to the use of
        # `torch.nonzero`.
        of_quality_inds = match_quality_matrix == highest_quality_foreach_gt[:, None]
        if of_quality_inds.dim() == 0:
            (_, pred_inds_with_highest_quality) = (
                of_quality_inds.unsqueeze(0).nonzero().unbind(1)
            )
        else:
            (_, pred_inds_with_highest_quality) = of_quality_inds.nonzero().unbind(1)
        match_labels[pred_inds_with_highest_quality] = 1


class RPNOutputs(object):
    def __init__(
        self,
        box2box_transform,
        anchor_matcher,
        batch_size_per_image,
        positive_fraction,
        images,
        pred_objectness_logits,
        pred_anchor_deltas,
        anchors,
        boundary_threshold=0,
        gt_boxes=None,
        smooth_l1_beta=0.0,
        ignorey=None,
        scales_yx=None
    ):
        """
        Args:
            box2box_transform (Box2BoxTransform): :class:`Box2BoxTransform` instance for anchor-proposal transformations.
            anchor_matcher (Matcher): :class:`Matcher` instance for matching anchors to ground-truth boxes; used to determine training labels.
            batch_size_per_image (int): number of proposals to sample when training
            positive_fraction (float): target fraction of sampled proposals that should be positive
            images (ImageList): :class:`ImageList` instance representing N input images
            pred_objectness_logits (list[Tensor]): A list of L elements. Element i is a tensor of shape (N, A, Hi, W)
            pred_anchor_deltas (list[Tensor]): A list of L elements. Element i is a tensor of shape (N, A*4, Hi, Wi)
            anchors (list[torch.Tensor]): nested list ofboxes. anchors[i][j] at (n, l) stores anchor array for feature map l
            boundary_threshold (int): if >= 0, then anchors that extend beyond the image boundary by more than boundary_thresh are not used in training.
            gt_boxes (list[Boxes], optional): A list of N elements.
            smooth_l1_beta (float): The transition point between L1 and L2 lossn. When set to 0, the loss becomes L1. When +inf, it is ignored
        """
        self.scales_yx=scales_yx
        self.box2box_transform = box2box_transform
        self.anchor_matcher = anchor_matcher
        self.batch_size_per_image = batch_size_per_image
        self.positive_fraction = positive_fraction
        self.pred_objectness_logits = pred_objectness_logits
        self.pred_anchor_deltas = pred_anchor_deltas
        self.ignorey = ignorey

        self.anchors = anchors
        self.gt_boxes = gt_boxes
        self.num_feature_maps = len(pred_objectness_logits)
        self.num_images = len(images)
        self.boundary_threshold = boundary_threshold
        self.smooth_l1_beta = smooth_l1_beta

    def _get_ground_truth(self):
        raise NotImplementedError()

    def predict_proposals(self):
        # pred_anchor_deltas: (L, N, ? Hi, Wi)
        # anchors:(N, L, -1, B)
        # here we loop over specific feature map, NOT images
        proposals = []
        anchors = self.anchors.transpose(0, 1)
        for anchors_i, pred_anchor_deltas_i in zip(anchors, self.pred_anchor_deltas):
            B = anchors_i.size(-1)
            N, _, Hi, Wi = pred_anchor_deltas_i.shape
            anchors_i = anchors_i.flatten(start_dim=0, end_dim=1)
            pred_anchor_deltas_i = (
                pred_anchor_deltas_i.view(N, -1, B, Hi, Wi)
                .permute(0, 3, 4, 1, 2)
                .reshape(-1, B)
            )
            proposals_i = self.box2box_transform.apply_deltas(
                pred_anchor_deltas_i, anchors_i, self.ignorey, self.scales_yx
            )
            # Append feature map proposals with shape (N, Hi*Wi*A, B)
            proposals.append(proposals_i.view(N, -1, B))
        proposals = torch.stack(proposals)
        return proposals

    def predict_objectness_logits(self):
        """
        Returns:
            pred_objectness_logits (list[Tensor]) -> (N, Hi*Wi*A).
        """
        pred_objectness_logits = [
            # Reshape: (N, A, Hi, Wi) -> (N, Hi, Wi, A) -> (N, Hi*Wi*A)
            score.permute(0, 2, 3, 1).reshape(self.num_images, -1)
            for score in self.pred_objectness_logits
        ]
        return pred_objectness_logits


# Main Classes
class Conv2d(torch.nn.Conv2d):
    def __init__(self, *args, **kwargs):
        norm = kwargs.pop("norm", None)
        activation = kwargs.pop("activation", None)
        super().__init__(*args, **kwargs)

        self.norm = norm
        self.activation = activation

    def forward(self, x):
        if x.numel() == 0 and self.training:
            assert not isinstance(self.norm, torch.nn.SyncBatchNorm)
        if x.numel() == 0:
            assert not isinstance(self.norm, torch.nn.GroupNorm)
            output_shape = [
                (i + 2 * p - (di * (k - 1) + 1)) // s + 1
                for i, p, di, k, s in zip(
                    x.shape[-2:],
                    self.padding,
                    self.dilation,
                    self.kernel_size,
                    self.stride,
                )
            ]
            output_shape = [x.shape[0], self.weight.shape[0]] + output_shape
            empty = _NewEmptyTensorOp.apply(x, output_shape)
            if self.training:
                _dummy = sum(x.view(-1)[0] for x in self.parameters()) * 0.0
                return empty + _dummy
            else:
                return empty

        x = super().forward(x)
        if self.norm is not None:
            x = self.norm(x)
        if self.activation is not None:
            x = self.activation(x)
        return x


class LastLevelMaxPool(nn.Module):
    """
    This module is used in the original FPN to generate a downsampled P6 feature from P5.
    """

    def __init__(self):
        super().__init__()
        self.num_levels = 1
        self.in_feature = "p5"

    def forward(self, x):
        return [F.max_pool2d(x, kernel_size=1, stride=2, padding=0)]


class LastLevelP6P7(nn.Module):
    """
    This module is used in RetinaNet to generate extra layers, P6 and P7 from C5 feature.
    """

    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.num_levels = 2
        self.in_feature = "res5"
        self.p6 = nn.Conv2d(in_channels, out_channels, 3, 2, 1)
        self.p7 = nn.Conv2d(out_channels, out_channels, 3, 2, 1)

    def forward(self, c5):
        p6 = self.p6(c5)
        p7 = self.p7(F.relu(p6))
        return [p6, p7]


class BasicStem(nn.Module):
    def __init__(self, in_channels=3, out_channels=64, norm="BN", caffe_maxpool=False):
        super().__init__()
        self.conv1 = Conv2d(
            in_channels,
            out_channels,
            kernel_size=7,
            stride=2,
            padding=3,
            bias=False,
            norm=get_norm(norm, out_channels),
        )
        self.caffe_maxpool = caffe_maxpool
        # use pad 1 instead of pad zero

    def forward(self, x):
        x = self.conv1(x)
        x = F.relu_(x)
        if self.caffe_maxpool:
            x = F.max_pool2d(x, kernel_size=3, stride=2, padding=0, ceil_mode=True)
        else:
            x = F.max_pool2d(x, kernel_size=3, stride=2, padding=1)
        return x

    @ property
    def out_channels(self):
        return self.conv1.out_channels

    @ property
    def stride(self):
        return 4  # = stride 2 conv -> stride 2 max pool


class ResNetBlockBase(nn.Module):
    def __init__(self, in_channels, out_channels, stride):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.stride = stride

    def freeze(self):
        for p in self.parameters():
            p.requires_grad = False
        return self


class BottleneckBlock(ResNetBlockBase):
    def __init__(
        self,
        in_channels,
        out_channels,
        bottleneck_channels,
        stride=1,
        num_groups=1,
        norm="BN",
        stride_in_1x1=False,
        dilation=1,
    ):
        super().__init__(in_channels, out_channels, stride)

        if in_channels != out_channels:
            self.shortcut = Conv2d(
                in_channels,
                out_channels,
                kernel_size=1,
                stride=stride,
                bias=False,
                norm=get_norm(norm, out_channels),
            )
        else:
            self.shortcut = None

        # The original MSRA ResNet models have stride in the first 1x1 conv
        # The subsequent fb.torch.resnet and Caffe2 ResNe[X]t implementations have
        # stride in the 3x3 conv
        stride_1x1, stride_3x3 = (stride, 1) if stride_in_1x1 else (1, stride)

        self.conv1 = Conv2d(
            in_channels,
            bottleneck_channels,
            kernel_size=1,
            stride=stride_1x1,
            bias=False,
            norm=get_norm(norm, bottleneck_channels),
        )

        self.conv2 = Conv2d(
            bottleneck_channels,
            bottleneck_channels,
            kernel_size=3,
            stride=stride_3x3,
            padding=1 * dilation,
            bias=False,
            groups=num_groups,
            dilation=dilation,
            norm=get_norm(norm, bottleneck_channels),
        )

        self.conv3 = Conv2d(
            bottleneck_channels,
            out_channels,
            kernel_size=1,
            bias=False,
            norm=get_norm(norm, out_channels),
        )

    def forward(self, x):
        out = self.conv1(x)
        out = F.relu_(out)

        out = self.conv2(out)
        out = F.relu_(out)

        out = self.conv3(out)

        if self.shortcut is not None:
            shortcut = self.shortcut(x)
        else:
            shortcut = x

        out += shortcut
        out = F.relu_(out)
        return out


class Backbone(nn.Module, metaclass=ABCMeta):
    def __init__(self):
        super().__init__()

    @ abstractmethod
    def forward(self):
        pass

    @ property
    def size_divisibility(self):
        """
        Some backbones require the input height and width to be divisible by a specific integer. This is
        typically true for encoder / decoder type networks with lateral connection (e.g., FPN) for which feature maps need to match
        dimension in the "bottom up" and "top down" paths. Set to 0 if no specific input size divisibility is required.
        """
        return 0

    def output_shape(self):
        return {
            name: ShapeSpec(
                channels=self._out_feature_channels[name],
                stride=self._out_feature_strides[name],
            )
            for name in self._out_features
        }

    @ property
    def out_features(self):
        """deprecated"""
        return self._out_features

    @ property
    def out_feature_strides(self):
        """deprecated"""
        return {f: self._out_feature_strides[f] for f in self._out_features}

    @ property
    def out_feature_channels(self):
        """deprecated"""
        return {f: self._out_feature_channels[f] for f in self._out_features}


class ResNet(Backbone):
    def __init__(self, stem, stages, num_classes=None, out_features=None):
        """
        Args:
            stem (nn.Module): a stem module
            stages (list[list[ResNetBlock]]): several (typically 4) stages, each contains multiple :class:`ResNetBlockBase`.
            num_classes (None or int): if None, will not perform classification.
            out_features (list[str]): name of the layers whose outputs should be returned in forward. Can be anything in:
            "stem", "linear", or "res2" ... If None, will return the output of the last layer.
        """
        super(ResNet, self).__init__()
        self.stem = stem
        self.num_classes = num_classes

        current_stride = self.stem.stride
        self._out_feature_strides = {"stem": current_stride}
        self._out_feature_channels = {"stem": self.stem.out_channels}

        self.stages_and_names = []
        for i, blocks in enumerate(stages):
            for block in blocks:
                assert isinstance(block, ResNetBlockBase), block
                curr_channels = block.out_channels
            stage = nn.Sequential(*blocks)
            name = "res" + str(i + 2)
            self.add_module(name, stage)
            self.stages_and_names.append((stage, name))
            self._out_feature_strides[name] = current_stride = int(
                current_stride * np.prod([k.stride for k in blocks])
            )
            self._out_feature_channels[name] = blocks[-1].out_channels

        if num_classes is not None:
            self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
            self.linear = nn.Linear(curr_channels, num_classes)

            # Sec 5.1 in "Accurate, Large Minibatch SGD: Training ImageNet in 1 Hour":
            # "The 1000-way fully-connected layer is initialized by
            # drawing weights from a zero-mean Gaussian with std of 0.01."
            nn.init.normal_(self.linear.weight, stddev=0.01)
            name = "linear"

        if out_features is None:
            out_features = [name]
        self._out_features = out_features
        assert len(self._out_features)
        children = [x[0] for x in self.named_children()]
        for out_feature in self._out_features:
            assert out_feature in children, "Available children: {}".format(
                ", ".join(children)
            )

    def forward(self, x):
        outputs = {}
        x = self.stem(x)
        if "stem" in self._out_features:
            outputs["stem"] = x
        for stage, name in self.stages_and_names:
            x = stage(x)
            if name in self._out_features:
                outputs[name] = x
        if self.num_classes is not None:
            x = self.avgpool(x)
            x = self.linear(x)
            if "linear" in self._out_features:
                outputs["linear"] = x
        return outputs

    def output_shape(self):
        return {
            name: ShapeSpec(
                channels=self._out_feature_channels[name],
                stride=self._out_feature_strides[name],
            )
            for name in self._out_features
        }

    @ staticmethod
    def make_stage(
        block_class,
        num_blocks,
        first_stride=None,
        *,
        in_channels,
        out_channels,
        **kwargs,
    ):
        """
        Usually, layers that produce the same feature map spatial size
        are defined as one "stage".
        Under such definition, stride_per_block[1:] should all be 1.
        """
        if first_stride is not None:
            assert "stride" not in kwargs and "stride_per_block" not in kwargs
            kwargs["stride_per_block"] = [first_stride] + [1] * (num_blocks - 1)
        blocks = []
        for i in range(num_blocks):
            curr_kwargs = {}
            for k, v in kwargs.items():
                if k.endswith("_per_block"):
                    assert len(v) == num_blocks, (
                        f"Argument '{k}' of make_stage should have the "
                        f"same length as num_blocks={num_blocks}."
                    )
                    newk = k[: -len("_per_block")]
                    assert (
                        newk not in kwargs
                    ), f"Cannot call make_stage with both {k} and {newk}!"
                    curr_kwargs[newk] = v[i]
                else:
                    curr_kwargs[k] = v

            blocks.append(
                block_class(
                    in_channels=in_channels, out_channels=out_channels, **curr_kwargs
                )
            )
            in_channels = out_channels

        return blocks


class ROIPooler(nn.Module):
    """
    Region of interest feature map pooler that supports pooling from one or more
    feature maps.
    """

    def __init__(
        self,
        output_size,
        scales,
        sampling_ratio,
        canonical_box_size=224,
        canonical_level=4,
    ):
        super().__init__()
        # assumption that stride is a power of 2.
        min_level = -math.log2(scales[0])
        max_level = -math.log2(scales[-1])

        # a bunch of testing
        assert math.isclose(min_level, int(min_level)) and math.isclose(max_level, int(max_level))
        assert (len(scales) == max_level - min_level + 1), "not pyramid"
        assert 0 < min_level and min_level <= max_level
        if isinstance(output_size, int):
            output_size = (output_size, output_size)
        assert len(output_size) == 2 and isinstance(output_size[0], int) and isinstance(output_size[1], int)
        if len(scales) > 1:
            assert (min_level <= canonical_level and canonical_level <= max_level)
        assert canonical_box_size > 0

        self.output_size = output_size
        self.min_level = int(min_level)
        self.max_level = int(max_level)
        self.level_poolers = nn.ModuleList(RoIPool(output_size, spatial_scale=scale) for scale in scales)
        self.canonical_level = canonical_level
        self.canonical_box_size = canonical_box_size

    def forward(self, feature_maps, boxes):
        """
        Args:
            feature_maps: List[torch.Tensor(N,C,W,H)]
            box_lists: list[torch.Tensor])
        Returns:
            A tensor of shape(N*B, Channels, output_size, output_size)
        """
        x = [v for v in feature_maps.values()]
        num_level_assignments = len(self.level_poolers)
        assert len(x) == num_level_assignments and len(boxes) == x[0].size(0)

        pooler_fmt_boxes = convert_boxes_to_pooler_format(boxes)

        if num_level_assignments == 1:
            return self.level_poolers[0](x[0], pooler_fmt_boxes)

        level_assignments = assign_boxes_to_levels(
            boxes,
            self.min_level,
            self.max_level,
            self.canonical_box_size,
            self.canonical_level,
        )

        num_boxes = len(pooler_fmt_boxes)
        num_channels = x[0].shape[1]
        output_size = self.output_size[0]

        dtype, device = x[0].dtype, x[0].device
        output = torch.zeros(
            (num_boxes, num_channels, output_size, output_size),
            dtype=dtype,
            device=device,
        )

        for level, (x_level, pooler) in enumerate(zip(x, self.level_poolers)):
            inds = torch.nonzero(level_assignments == level).squeeze(1)
            pooler_fmt_boxes_level = pooler_fmt_boxes[inds]
            output[inds] = pooler(x_level, pooler_fmt_boxes_level)

        return output


class ROIOutputs(object):

    def __init__(self, cfg, training=False):
        self.smooth_l1_beta = cfg.ROI_BOX_HEAD.SMOOTH_L1_BETA
        self.box2box_transform = Box2BoxTransform(weights=cfg.ROI_BOX_HEAD.BBOX_REG_WEIGHTS)
        self.training = training
        self.score_thresh = cfg.ROI_HEADS.SCORE_THRESH_TEST
        self.min_detections = cfg.MIN_DETECTIONS
        self.max_detections = cfg.MAX_DETECTIONS

        nms_thresh = cfg.ROI_HEADS.NMS_THRESH_TEST
        if not isinstance(nms_thresh, list):
            nms_thresh = [nms_thresh]
        self.nms_thresh = nms_thresh

    def _predict_boxes(self, proposals, box_deltas, preds_per_image):
        num_pred = box_deltas.size(0)
        B = proposals[0].size(-1)
        K = box_deltas.size(-1) // B
        box_deltas = box_deltas.view(num_pred * K, B)
        proposals = torch.cat(proposals, dim=0).unsqueeze(-2).expand(num_pred, K, B)
        proposals = proposals.reshape(-1, B)
        boxes = self.box2box_transform.apply_deltas(box_deltas, proposals)
        return boxes.view(num_pred, K * B).split(preds_per_image, dim=0)

    def _predict_objs(self, obj_logits, preds_per_image):
        probs = F.softmax(obj_logits, dim=-1)
        probs = probs.split(preds_per_image, dim=0)
        return probs

    def _predict_attrs(self, attr_logits, preds_per_image):
        attr_logits = attr_logits[..., :-1].softmax(-1)
        attr_probs, attrs = attr_logits.max(-1)
        return attr_probs.split(preds_per_image, dim=0), attrs.split(preds_per_image, dim=0)

    @torch.no_grad()
    def inference(self, obj_logits, attr_logits, box_deltas, pred_boxes, features, sizes, scales=None):
        # only the pred boxes is the
        preds_per_image = [p.size(0) for p in pred_boxes]
        boxes_all = self._predict_boxes(pred_boxes, box_deltas, preds_per_image)
        obj_scores_all = self._predict_objs(obj_logits, preds_per_image)  # list of length N
        attr_probs_all, attrs_all = self._predict_attrs(attr_logits, preds_per_image)
        features = features.split(preds_per_image, dim=0)

        final_results = []
        zipped = zip(boxes_all, obj_scores_all, attr_probs_all, attrs_all, sizes)
        for i, (boxes, obj_scores, attr_probs, attrs, size) in enumerate(zipped):
            for nms_t in self.nms_thresh:
                outputs = do_nms(boxes, obj_scores, size, self.score_thresh, nms_t, self.min_detections, self.max_detections)
                stop, max_boxes, max_scores, classes, ids = outputs
                if stop:
                    break

            if scales is not None:
                scale_yx = scales[i]
                max_boxes[:, 0::2] *= scale_yx[1]
                max_boxes[:, 1::2] *= scale_yx[0]

            final_results.append((
                max_boxes,
                classes,
                max_scores,
                attrs[ids],
                attr_probs[ids],
                features[i][ids]
            ))
        boxes, classes, class_probs, attrs, attr_probs, roi_features = map(list, zip(*final_results))
        return boxes, classes, class_probs, attrs, attr_probs, roi_features

    def training(self, obj_logits, attr_logits, box_deltas, pred_boxes, features, sizes):
        pass

    def __call__(self, obj_logits, attr_logits, box_deltas, pred_boxes, features, sizes, scales=None):
        if self.training:
            raise NotImplementedError()
        return self.inference(obj_logits, attr_logits, box_deltas, pred_boxes, features, sizes, scales=scales)


class Res5ROIHeads(nn.Module):
    """
    ROIHeads perform all per-region computation in an R-CNN.
    It contains logic of cropping the regions, extract per-region features
    (by the res-5 block in this case), and make per-region predictions.
    """

    def __init__(self, cfg, input_shape):
        super().__init__()
        self.batch_size_per_image = cfg.RPN.BATCH_SIZE_PER_IMAGE
        self.positive_sample_fraction = cfg.ROI_HEADS.POSITIVE_FRACTION
        self.in_features = cfg.ROI_HEADS.IN_FEATURES
        self.num_classes = cfg.ROI_HEADS.NUM_CLASSES
        self.proposal_append_gt = cfg.ROI_HEADS.PROPOSAL_APPEND_GT
        self.feature_strides = {k: v.stride for k, v in input_shape.items()}
        self.feature_channels = {k: v.channels for k, v in input_shape.items()}
        self.cls_agnostic_bbox_reg = cfg.ROI_BOX_HEAD.CLS_AGNOSTIC_BBOX_REG
        self.stage_channel_factor = 2 ** 3  # res5 is 8x res2
        self.out_channels = cfg.RESNETS.RES2_OUT_CHANNELS * self.stage_channel_factor

        # self.proposal_matcher = Matcher(
        #     cfg.ROI_HEADS.IOU_THRESHOLDS,
        #     cfg.ROI_HEADS.IOU_LABELS,
        #     allow_low_quality_matches=False,
        # )

        pooler_resolution = cfg.ROI_BOX_HEAD.POOLER_RESOLUTION
        pooler_scales = (1.0 / self.feature_strides[self.in_features[0]], )
        sampling_ratio = cfg.ROI_BOX_HEAD.POOLER_SAMPLING_RATIO
        res5_halve = cfg.ROI_BOX_HEAD.RES5HALVE
        use_attr = cfg.ROI_BOX_HEAD.ATTR
        num_attrs = cfg.ROI_BOX_HEAD.NUM_ATTRS

        self.pooler = ROIPooler(
            output_size=pooler_resolution,
            scales=pooler_scales,
            sampling_ratio=sampling_ratio,
        )

        self.res5 = self._build_res5_block(cfg)
        if not res5_halve:
            """
            Modifications for VG in RoI heads:
            1. Change the stride of conv1 and shortcut in Res5.Block1 from 2 to 1
            2. Modifying all conv2 with (padding: 1 --> 2) and (dilation: 1 --> 2)
            """
            self.res5[0].conv1.stride = (1, 1)
            self.res5[0].shortcut.stride = (1, 1)
            for i in range(3):
                self.res5[i].conv2.padding = (2, 2)
                self.res5[i].conv2.dilation = (2, 2)

        self.box_predictor = FastRCNNOutputLayers(
            self.out_channels,
            self.num_classes,
            self.cls_agnostic_bbox_reg,
            use_attr=use_attr,
            num_attrs=num_attrs,
        )

    def _build_res5_block(self, cfg):
        stage_channel_factor = self.stage_channel_factor  # res5 is 8x res2
        num_groups = cfg.RESNETS.NUM_GROUPS
        width_per_group = cfg.RESNETS.WIDTH_PER_GROUP
        bottleneck_channels = num_groups * width_per_group * stage_channel_factor
        out_channels = self.out_channels
        stride_in_1x1 = cfg.RESNETS.STRIDE_IN_1X1
        norm = cfg.RESNETS.NORM

        blocks = ResNet.make_stage(
            BottleneckBlock,
            3,
            first_stride=2,
            in_channels=out_channels // 2,
            bottleneck_channels=bottleneck_channels,
            out_channels=out_channels,
            num_groups=num_groups,
            norm=norm,
            stride_in_1x1=stride_in_1x1,
        )
        return nn.Sequential(*blocks)

    def _shared_roi_transform(self, features, boxes):
        x = self.pooler(features, boxes)
        return self.res5(x)

    def forward(self, features, proposal_boxes, gt_boxes=None):
        if self.training:
            """
            see https://github.com/airsplay/py-bottom-up-attention/\
                    blob/master/detectron2/modeling/roi_heads/roi_heads.py
            """
            raise NotImplementedError()

        assert not proposal_boxes[0].requires_grad
        box_features = self._shared_roi_transform(features, proposal_boxes)
        feature_pooled = box_features.mean(dim=[2, 3])  # pooled to 1x1
        obj_logits, attr_logits, pred_proposal_deltas = self.box_predictor(feature_pooled)
        return obj_logits, attr_logits, pred_proposal_deltas, feature_pooled


class AnchorGenerator(nn.Module):
    """
    For a set of image sizes and feature maps, computes a set of anchors.
    """

    def __init__(self, cfg, input_shape: List[ShapeSpec]):
        super().__init__()
        # fmt: off
        sizes = cfg.ANCHOR_GENERATOR.SIZES
        aspect_ratios = cfg.ANCHOR_GENERATOR.ASPECT_RATIOS
        self.strides = [x.stride for x in input_shape]
        self.offset = cfg.ANCHOR_GENERATOR.OFFSET
        assert 0.0 <= self.offset < 1.0, self.offset

        """
        sizes (list[list[int]]): sizes[i] is the list of anchor sizes for feat map i
            1. given in absolute lengths in units of the input image;
            2. they do not dynamically scale if the input image size changes.
        aspect_ratios (list[list[float]])
        strides (list[int]): stride of each input feature.
        """

        self.num_features = len(self.strides)
        self.cell_anchors = nn.ParameterList(
            self._calculate_anchors(sizes, aspect_ratios)
        )
        self._spacial_feat_dim = 4

    def _calculate_anchors(self, sizes, aspect_ratios):
        # If one size (or aspect ratio) is specified and there are multiple feature
        # maps, then we "broadcast" anchors of that single size (or aspect ratio)
        if len(sizes) == 1:
            sizes *= self.num_features
        if len(aspect_ratios) == 1:
            aspect_ratios *= self.num_features
        assert self.num_features == len(sizes)
        assert self.num_features == len(aspect_ratios)

        cell_anchors = [
            self.generate_cell_anchors(s, a).float()
            for s, a in zip(sizes, aspect_ratios)
        ]

        return cell_anchors

    @property
    def box_dim(self):
        return self._spacial_feat_dim

    @property
    def num_cell_anchors(self):
        """
        Returns:
            list[int]: Each int is the number of anchors at every pixel location, on that feature map.
        """
        return [len(cell_anchors) for cell_anchors in self.cell_anchors]

    def grid_anchors(self, grid_sizes):
        anchors = []
        for (size, stride, base_anchors) in zip(
            grid_sizes, self.strides, self.cell_anchors
        ):
            shift_x, shift_y = _create_grid_offsets(
                size, stride, self.offset, base_anchors.device
            )
            shifts = torch.stack((shift_x, shift_y, shift_x, shift_y), dim=1)

            anchors.append(
                (shifts.view(-1, 1, 4) + base_anchors.view(1, -1, 4)).reshape(-1, 4)
            )

        return anchors

    def generate_cell_anchors(
        self, sizes=(32, 64, 128, 256, 512), aspect_ratios=(0.5, 1, 2)
    ):
        """
        anchors are continious geometric rectangles
        centered on one feature map point sample.
        We can later build the set of anchors
        for the entire feature map by tiling these tensors
        """

        anchors = []
        for size in sizes:
            area = size ** 2.0
            for aspect_ratio in aspect_ratios:
                w = math.sqrt(area / aspect_ratio)
                h = aspect_ratio * w
                x0, y0, x1, y1 = -w / 2.0, -h / 2.0, w / 2.0, h / 2.0
                anchors.append([x0, y0, x1, y1])
        return nn.Parameter(torch.Tensor(anchors))

    def forward(self, features):
        """
        Args:
            features List[torch.Tensor]: list of feature maps on which to generate anchors.
        Returns:
            torch.Tensor: a list of #image elements.
        """
        num_images = features[0].size(0)
        grid_sizes = [feature_map.shape[-2:] for feature_map in features]
        anchors_over_all_feature_maps = self.grid_anchors(grid_sizes)
        anchors_over_all_feature_maps = torch.stack(anchors_over_all_feature_maps)
        return anchors_over_all_feature_maps.unsqueeze(0).repeat_interleave(num_images, dim=0)


class RPNHead(nn.Module):
    """
    RPN classification and regression heads. Uses a 3x3 conv to produce a shared
    hidden state from which one 1x1 conv predicts objectness logits for each anchor
    and a second 1x1 conv predicts bounding-box deltas specifying how to deform
    each anchor into an object proposal.
    """

    def __init__(self, cfg, input_shape: List[ShapeSpec]):
        super().__init__()

        # Standard RPN is shared across levels:
        in_channels = [s.channels for s in input_shape]
        assert len(set(in_channels)) == 1, "Each level must have the same channel!"
        in_channels = in_channels[0]

        anchor_generator = AnchorGenerator(cfg, input_shape)
        num_cell_anchors = anchor_generator.num_cell_anchors
        box_dim = anchor_generator.box_dim
        assert (
            len(set(num_cell_anchors)) == 1
        ), "Each level must have the same number of cell anchors"
        num_cell_anchors = num_cell_anchors[0]

        if cfg.PROPOSAL_GENERATOR.HIDDEN_CHANNELS == -1:
            hid_channels = in_channels
        else:
            hid_channels = cfg.PROPOSAL_GENERATOR.HIDDEN_CHANNELS
            # Modifications for VG in RPN (modeling/proposal_generator/rpn.py)
            # Use hidden dim  instead fo the same dim as Res4 (in_channels)

        # 3x3 conv for the hidden representation
        self.conv = nn.Conv2d(
            in_channels, hid_channels, kernel_size=3, stride=1, padding=1
        )
        # 1x1 conv for predicting objectness logits
        self.objectness_logits = nn.Conv2d(
            hid_channels, num_cell_anchors, kernel_size=1, stride=1
        )
        # 1x1 conv for predicting box2box transform deltas
        self.anchor_deltas = nn.Conv2d(
            hid_channels, num_cell_anchors * box_dim, kernel_size=1, stride=1
        )

        for layer in [self.conv, self.objectness_logits, self.anchor_deltas]:
            nn.init.normal_(layer.weight, std=0.01)
            nn.init.constant_(layer.bias, 0)

    def forward(self, features):
        """
        Args:
            features (list[Tensor]): list of feature maps
        """
        pred_objectness_logits = []
        pred_anchor_deltas = []
        for x in features:
            t = F.relu(self.conv(x))
            pred_objectness_logits.append(self.objectness_logits(t))
            pred_anchor_deltas.append(self.anchor_deltas(t))
        return pred_objectness_logits, pred_anchor_deltas


class RPN(nn.Module):
    """
    Region Proposal Network, introduced by the Faster R-CNN paper.
    """

    def __init__(self, cfg, input_shape: Dict[str, ShapeSpec]):
        super().__init__()

        self.min_box_side_len = cfg.PROPOSAL_GENERATOR.MIN_SIZE
        self.in_features = cfg.RPN.IN_FEATURES
        self.nms_thresh = cfg.RPN.NMS_THRESH
        self.batch_size_per_image = cfg.RPN.BATCH_SIZE_PER_IMAGE
        self.positive_fraction = cfg.RPN.POSITIVE_FRACTION
        self.smooth_l1_beta = cfg.RPN.SMOOTH_L1_BETA
        self.loss_weight = cfg.RPN.LOSS_WEIGHT

        self.pre_nms_topk = {
            True: cfg.RPN.PRE_NMS_TOPK_TRAIN,
            False: cfg.RPN.PRE_NMS_TOPK_TEST,
        }
        self.post_nms_topk = {
            True: cfg.RPN.POST_NMS_TOPK_TRAIN,
            False: cfg.RPN.POST_NMS_TOPK_TEST,
        }
        self.boundary_threshold = cfg.RPN.BOUNDARY_THRESH

        self.anchor_generator = AnchorGenerator(
            cfg, [input_shape[f] for f in self.in_features]
        )
        self.box2box_transform = Box2BoxTransform(weights=cfg.RPN.BBOX_REG_WEIGHTS)
        self.anchor_matcher = Matcher(
            cfg.RPN.IOU_THRESHOLDS,
            cfg.RPN.IOU_LABELS,
            allow_low_quality_matches=True,
        )
        self.rpn_head = RPNHead(cfg, [input_shape[f] for f in self.in_features])

    def training(self, images, image_shapes, features, gt_boxes):
        pass

    def inference(self, outputs, images, image_shapes, features, gt_boxes=None, scales_yx=None, ignorey=None):
        outputs = find_top_rpn_proposals(
            outputs.predict_proposals(),
            outputs.predict_objectness_logits(),
            images,
            image_shapes,
            self.nms_thresh,
            self.pre_nms_topk[self.training],
            self.post_nms_topk[self.training],
            self.min_box_side_len,
            self.training,
            scales_yx=scales_yx,
            ignorey=ignorey
        )

        results = []
        for img in outputs:
            im_boxes, img_box_logits = img
            img_box_logits, inds = img_box_logits.sort(descending=True)
            im_boxes = im_boxes[inds]
            results.append((im_boxes, img_box_logits))

        (proposal_boxes, logits) = tuple(map(list, zip(*results)))
        return proposal_boxes, logits

    def forward(self, images, image_shapes, features, gt_boxes=None, ignorey=None, scales_yx=None):
        """
        Args:
            images (torch.Tensor): input images of length `N`
            features (dict[str: Tensor])
            gt_instances
        """
        # features is dict, key = block level, v = feature_map
        features = [features[f] for f in self.in_features]
        pred_objectness_logits, pred_anchor_deltas = self.rpn_head(features)
        anchors = self.anchor_generator(features)
        outputs = RPNOutputs(
            self.box2box_transform,
            self.anchor_matcher,
            self.batch_size_per_image,
            self.positive_fraction,
            images,
            pred_objectness_logits,
            pred_anchor_deltas,
            anchors,
            self.boundary_threshold,
            gt_boxes,
            self.smooth_l1_beta,
            ignorey=ignorey,
            scales_yx=scales_yx
        )
        # For RPN-only models, the proposals are the final output

        if self.training:
            raise NotImplementedError()
            return self.training(outputs, images, image_shapes, features, gt_boxes)
        else:
            return self.inference(outputs, images, image_shapes, features, gt_boxes,
                    ignorey=ignorey, scales_yx=scales_yx)


class FastRCNNOutputLayers(nn.Module):
    """
    Two linear layers for predicting Fast R-CNN outputs:
      (1) proposal-to-detection box regression deltas
      (2) classification scores
    """

    def __init__(
        self,
        input_size,
        num_classes,
        cls_agnostic_bbox_reg,
        box_dim=4,
        use_attr=False,
        num_attrs=-1,
    ):
        """
        Args:
            input_size (int): channels, or (channels, height, width)
            num_classes (int)
            cls_agnostic_bbox_reg (bool)
            box_dim (int)
        """
        super().__init__()

        if not isinstance(input_size, int):
            input_size = np.prod(input_size)

        # (do + 1 for background class)
        self.cls_score = nn.Linear(input_size, num_classes + 1)
        num_bbox_reg_classes = 1 if cls_agnostic_bbox_reg else num_classes
        self.bbox_pred = nn.Linear(input_size, num_bbox_reg_classes * box_dim)

        self.use_attr = use_attr
        if use_attr:
            '''
            Modifications for VG in RoI heads
            Embedding: {num_classes + 1} --> {input_size // 8}
            Linear: {input_size + input_size // 8} --> {input_size // 4}
            Linear: {input_size // 4} --> {num_attrs + 1}
            '''
            self.cls_embedding = nn.Embedding(num_classes + 1, input_size // 8)
            self.fc_attr = nn.Linear(input_size + input_size // 8, input_size // 4)
            self.attr_score = nn.Linear(input_size // 4, num_attrs + 1)

        nn.init.normal_(self.cls_score.weight, std=0.01)
        nn.init.normal_(self.bbox_pred.weight, std=0.001)
        for item in [self.cls_score, self.bbox_pred]:
            nn.init.constant_(item.bias, 0)

    def forward(self, roi_features):
        if roi_features.dim() > 2:
            roi_features = torch.flatten(roi_features, start_dim=1)
        scores = self.cls_score(roi_features)
        proposal_deltas = self.bbox_pred(roi_features)
        if self.use_attr:
            _, max_class = scores.max(-1)  # [b, c] --> [b]
            cls_emb = self.cls_embedding(max_class)  # [b] --> [b, 256]
            roi_features = torch.cat([roi_features, cls_emb], -1)  # [b, 2048] + [b, 256] --> [b, 2304]
            roi_features = self.fc_attr(roi_features)
            roi_features = F.relu(roi_features)
            attr_scores = self.attr_score(roi_features)
            return scores, attr_scores, proposal_deltas
        else:
            return scores, proposal_deltas


class FRCNN(nn.Module):
    def __init__(self, cfg):
        super().__init__()

        self.min_detections = cfg.min_detections
        self.max_detections = cfg.max_detections
        self.config = cfg
        self.device = torch.device(cfg.MODEL.DEVICE)
        self.backbone = build_backbone(cfg)
        self.proposal_generator = RPN(cfg, self.backbone.output_shape())
        self.roi_heads = Res5ROIHeads(cfg, self.backbone.output_shape())
        self.roi_outputs = ROIOutputs(cfg)
        self.to(cfg.model.device)

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, *model_args, **kwargs):
        config = kwargs.pop("config", None)
        state_dict = kwargs.pop("state_dict", None)
        cache_dir = kwargs.pop("cache_dir", None)
        from_tf = kwargs.pop("from_tf", False)
        force_download = kwargs.pop("force_download", False)
        resume_download = kwargs.pop("resume_download", False)
        proxies = kwargs.pop("proxies", None)
        local_files_only = kwargs.pop("local_files_only", False)
        use_cdn = kwargs.pop("use_cdn", True)

        # Load config if we don't provide a configuration
        if not isinstance(config, Config):
            config_path = config if config is not None else pretrained_model_name_or_path
            # try:
            config = Config.from_pretrained(
                config_path,
                cache_dir=cache_dir,
                force_download=force_download,
                resume_download=resume_download,
                proxies=proxies,
                local_files_only=local_files_only,
            )

        # Load model
        if pretrained_model_name_or_path is not None:
            if os.path.isdir(pretrained_model_name_or_path):
                if os.path.isfile(os.path.join(pretrained_model_name_or_path, WEIGHTS_NAME)):
                    # Load from a PyTorch checkpoint
                    archive_file = os.path.join(pretrained_model_name_or_path, WEIGHTS_NAME)
                else:
                    raise EnvironmentError(
                        "Error no file named {} found in directory {} ".format(
                            WEIGHTS_NAME,
                            pretrained_model_name_or_path,
                        )
                    )
            elif os.path.isfile(pretrained_model_name_or_path) or is_remote_url(pretrained_model_name_or_path):
                archive_file = pretrained_model_name_or_path
            elif os.path.isfile(pretrained_model_name_or_path + ".index"):
                assert (
                    from_tf
                ), "We found a TensorFlow checkpoint at {}, please set from_tf to True to load from this checkpoint".format(
                    pretrained_model_name_or_path + ".index"
                )
                archive_file = pretrained_model_name_or_path + ".index"
            else:
                archive_file = hf_bucket_url(
                    pretrained_model_name_or_path,
                    filename=WEIGHTS_NAME,
                    use_cdn=use_cdn,
                )

            try:
                # Load from URL or cache if already cached
                resolved_archive_file = cached_path(
                    archive_file,
                    cache_dir=cache_dir,
                    force_download=force_download,
                    proxies=proxies,
                    resume_download=resume_download,
                    local_files_only=local_files_only,
                )
                if resolved_archive_file is None:
                    raise EnvironmentError
            except EnvironmentError:
                msg = (
                    f"Can't load weights for '{pretrained_model_name_or_path}'."
                )
                raise EnvironmentError(msg)

            if resolved_archive_file == archive_file:
                print("loading weights file {}".format(archive_file))
            else:
                print("loading weights file {} from cache at {}".format(archive_file, resolved_archive_file))
        else:
            resolved_archive_file = None


        #config.device = "cuda:0"
        model = cls(config)
        model = model.to(torch.device(config.model.device))
        #model.cuda()

        if state_dict is None:
            try:
                try:
                    state_dict = torch.load(resolved_archive_file, map_location=torch.device(config.model.device))
                except Exception:
                    state_dict = load_checkpoint(resolved_archive_file)

            except Exception:
                raise OSError(
                    "Unable to load weights from pytorch checkpoint file. "
                    "If you tried to load a PyTorch model from a TF 2.0 checkpoint, please set from_tf=True. "
                )

        missing_keys = []
        unexpected_keys = []
        error_msgs = []

        # Convert old format to new format if needed from a PyTorch state_dict
        old_keys = []
        new_keys = []
        for key in state_dict.keys():
            new_key = None
            if "gamma" in key:
                new_key = key.replace("gamma", "weight")
            if "beta" in key:
                new_key = key.replace("beta", "bias")
            if new_key:
                old_keys.append(key)
                new_keys.append(new_key)
        for old_key, new_key in zip(old_keys, new_keys):
            state_dict[new_key] = state_dict.pop(old_key)

        # copy state_dict so _load_from_state_dict can modify it
        metadata = getattr(state_dict, "_metadata", None)
        state_dict = state_dict.copy()
        if metadata is not None:
            state_dict._metadata = metadata

        model_to_load = model
        model_to_load.load_state_dict(state_dict)

        if model.__class__.__name__ != model_to_load.__class__.__name__:
            base_model_state_dict = model_to_load.state_dict().keys()
            head_model_state_dict_without_base_prefix = [
                key.split(cls.base_model_prefix + ".")[-1] for key in model.state_dict().keys()
            ]
            missing_keys.extend(head_model_state_dict_without_base_prefix - base_model_state_dict)

        if len(unexpected_keys) > 0:
            print(
                f"Some weights of the model checkpoint at {pretrained_model_name_or_path} were not used when "
                f"initializing {model.__class__.__name__}: {unexpected_keys}\n"
                f"- This IS expected if you are initializing {model.__class__.__name__} from the checkpoint of a model trained on another task "
                f"or with another architecture (e.g. initializing a BertForSequenceClassification model from a BertForPretraining model).\n"
                f"- This IS NOT expected if you are initializing {model.__class__.__name__} from the checkpoint of a model that you expect "
                f"to be exactly identical (initializing a BertForSequenceClassification model from a BertForSequenceClassification model)."
            )
        else:
            print(f"All model checkpoint weights were used when initializing {model.__class__.__name__}.\n")
        if len(missing_keys) > 0:
            print(
                f"Some weights of {model.__class__.__name__} were not initialized from the model checkpoint at {pretrained_model_name_or_path} "
                f"and are newly initialized: {missing_keys}\n"
                f"You should probably TRAIN this model on a down-stream task to be able to use it for predictions and inference."
            )
        else:
            print(
                f"All the weights of {model.__class__.__name__} were initialized from the model checkpoint at {pretrained_model_name_or_path}.\n"
                f"If your task is similar to the task the model of the checkpoint was trained on, "
                f"you can already use {model.__class__.__name__} for predictions without further training."
            )
        if len(error_msgs) > 0:
            raise RuntimeError(
                "Error(s) in loading state_dict for {}:\n\t{}".format(
                    model.__class__.__name__, "\n\t".join(error_msgs)
                )
            )
        # Set model in evaluation mode to deactivate DropOut modules by default
        model.eval()

        return model

    def forward(self, images, image_shapes, gt_boxes=None, proposals=None, scales_yx=None, ignorey=None,**kwargs):
        '''
        kwargs:
            max_detections (int), return_tensors {"np", "pt", None}, padding {None,
            "max_detections"}, pad_value (int), location = {"cuda", "cpu"}
        '''
        if self.training:
            raise NotImplementedError()
        return self.inference(
            images=images,
            image_shapes=image_shapes,
            gt_boxes=gt_boxes,
            proposals=proposals,
            scales_yx=scales_yx,
            ignorey=ignorey,
            **kwargs
        )

    @torch.no_grad()
    def inference(self, images, image_shapes, gt_boxes=None, proposals=None, scales_yx=None, ignorey=None, **kwargs):
        # run images through bacbone
        features = self.backbone(images)

        # generate proposals if none are available
        if proposals is None:
            proposal_boxes, _ = self.proposal_generator(
                images,
                image_shapes,
                features,
                gt_boxes,
                ignorey,
                scales_yx
            )
        else:
            assert proposals is not None

        # pool object features from either gt_boxes, or from proposals
        obj_logits, attr_logits, box_deltas, feature_pooled = self.roi_heads(
            features,
            proposal_boxes,
            gt_boxes
        )

        # prepare FRCNN Outputs and select top proposals
        boxes, classes, class_probs, attrs, attr_probs, roi_features = self.roi_outputs(
            obj_logits=obj_logits,
            attr_logits=attr_logits,
            box_deltas=box_deltas,
            pred_boxes=proposal_boxes,
            features=feature_pooled,
            sizes=image_shapes,
            scales=scales_yx
        )

        # will we pad???
        # subset_kwargs = {
        #     "max_detections": self.max_detections,
        #     "return_tensors": kwargs.get("return_tensors", None),
        #     "pad_value": kwargs.get("pad_value", 0),
        #     "padding": kwargs.get("padding", None)
        # }
        preds_per_image = torch.tensor([p.size(0) for p in boxes])
        # boxes = pad_list_tensors(boxes, preds_per_image, **subset_kwargs)
        # classes = pad_list_tensors(classes, preds_per_image, **subset_kwargs)
        # class_probs = pad_list_tensors(class_probs, preds_per_image, **subset_kwargs)
        # attrs = pad_list_tensors(attrs, preds_per_image, **subset_kwargs)
        # attr_probs = pad_list_tensors(attr_probs, preds_per_image, **subset_kwargs)
        # roi_features = pad_list_tensors(roi_features, preds_per_image, **subset_kwargs)
        # subset_kwargs["padding"] = None
        # preds_per_image = pad_list_tensors(preds_per_image, None, **subset_kwargs)
        # sizes = pad_list_tensors(image_shapes, None, **subset_kwargs)
        # normalized_boxes = norm_box(boxes, original_sizes)
        return OrderedDict({
            "obj_ids": classes,
            "obj_probs": class_probs,
            "attr_ids": attrs,
            "attr_probs": attr_probs,
            "boxes": boxes,
            "preds_per_image": preds_per_image,
            "roi_features": roi_features,
        })



import vltk.vars as vltk
from tqdm import tqdm
from vltk import adapters


class COCOCaptions(adapters.VisnLangDataset):
    data_info = {
        "train": {"coco2014": ["train"]},
        "val": {"coco2014": ["val"]},
    }

    @staticmethod
    def schema():
        return {}

    @staticmethod
    def forward(json_files, split, min_label_frequency=2):
        batch_entries = []
        id2imgid = {}
        for filename, data in json_files.items():
            if "annotations" not in data:
                continue
            if "caption" not in data["annotations"][0]:
                continue
            for img in data["images"]:
                id2imgid[img["id"]] = img["file_name"]
            for item in tqdm(data["annotations"]):
                imgid = id2imgid[item["image_id"]].split(".")[0]
                entry = {vltk.imgid: imgid, vltk.text: item["caption"]}

                batch_entries.append(entry)

        return batch_entries



import os

"""
ALL BOXES ARE EXPECTED TO GO IN: (X, Y, W, H) FORMAT
"""

# pathes
ANNOTATION_DIR = "annotations"
BASEPATH = os.path.abspath(os.path.join(os.path.dirname(__file__), "."))
VISIONLANG = os.path.join(BASEPATH, "adapters/visionlang")
VISION = os.path.join(BASEPATH, "adapters/vision")
EXTRACTION = os.path.join(BASEPATH, "adapters/vision")
ADAPTERS = os.path.join(BASEPATH, "adapters")
PROCESSORS = os.path.join(BASEPATH, "processing")
SIMPLEPATH = os.path.join(BASEPATH, "experiment")
MODELPATH = os.path.join(BASEPATH, "modeling")
LOOPPATH = os.path.join(BASEPATH, "loop")
SCHEDPATH = os.path.join(BASEPATH, "processing/sched.py")
DATAPATH = os.path.join(BASEPATH, "processing/data.py")
LABELPROCPATH = os.path.join(BASEPATH, "processing/label.py")
IMAGEPROCPATH = os.path.join(BASEPATH, "processing/image.py")
OPTIMPATH = os.path.join(BASEPATH, "processing/optim.py")
VOCABPATH = os.path.join(BASEPATH, "libdata", "tokenizer_files", "vocab.txt")

# special deliminator
delim = "^"
tokenmap = "tokenmap"
# common keys across library
span = "span"
n_objects = "n_objects"
objects = "objects"
type_ids = "type_ids"
input_ids = "input_ids"
tokenboxes = "tokenboxes"
tokenbox = "tokenbox"
text_attention_mask = "text_attention_mask"
rawsize = "rawsize"
padsize = "padsize"
size = "size"
polygons = "poly"
RLE = "RLE"
segmentations = "segmentations"
segmentation = "segmentation"  # legacy
boxes = "boxes"
box = "box"  # legacy
imgid = "imgid"
labels = "labels"
label = "label"
text = "text"
scores = "scores"
score = "score"
img = "image"
filepath = "filepath"
features = "features"
split = "split"
scale = "wh_scale"
boxtensor = "boxtensor"
area = "area"
size = "size"
qid = "qid"

SPLITALIASES = {
    "test",
    "dev",
    "eval",
    "val",
    "validation",
    "evaluation",
    "train",
}


VLOVERLAP = {
    text: "vtext",
    labels: "vlabels",
    label: "vlabel",
    scores: "vscores",
    score: "vscore",
}


# dataset selection values
VLDATA = 0
VDATA = 1
LDATA = 2


SUPPORTEDNAMES = {
    type_ids,
    input_ids,
    text_attention_mask,
    rawsize,
    size,
    segmentation,
    box,
    imgid,
    label,
    text,
    score,
    label,
    text,
    score,
    img,
    filepath,
    features,
    split,
    scale,
    boxtensor,
    area,
    size,
}



import torch
from vltk.modeling.botnet_v1 import BottleStack
from vltk.modeling.botnet_v2 import ResNet50
from vltk.modeling.botnet_v3 import MyBotNet
from vltk.utils import get_nvidia_gpu_memory

scaler = torch.cuda.amp.GradScaler()

size = 768
bz = 4

print("init model")
# model1 = BottleStack(dim=3, fmap_size=size)
model2 = MyBotNet(resolution=(size, size))
model2.to(torch.device(3))
# print("Bottle Stack")
# for n, p in model1.named_parameters():
#     print(n, p.shape)
# print("resent")
# for n, p in model2.named_parameters():
#     print(n, p.shape)


# print("data")
x = torch.rand(bz, 3, size, size)
x = x.to(torch.device(3))

with torch.cuda.amp.autocast():
    output = model2(x)
    print(output.shape)



# FOR CODE BORROWED FROM VISUALTRANSFORMER RELEASE TO CONVERT CHECKPOINTS:
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import collections
import os

import ml_collections
import numpy as np
import torch
from vltk.configs import Config, ModelConfig, PathesConfig, TrainConfig
from vltk.models.vit import VisionTransformer as VisionTransformerPytorch
from vltk. import convert_jax_to_torch_weights

# CONFIGS, KNOWN_MODELS,

# from vit_checkpoint


def _flatten_dict(d, parent_key="", sep="/"):
    """Flattens a dictionary, keeping empty leaves."""
    items = []
    for k, v in d.items():
        path = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(_flatten_dict(v, path, sep=sep).items())
        else:
            items.append((path, v))

    # Keeps the empty dict if it was set explicitly.
    if parent_key and not d:
        items.append((parent_key, {}))

    return dict(items)


def inspect_params(
    *, params, expected, logger, fail_if_extra=True, fail_if_missing=True
):
    """Inspects whether the params are consistent with the expected keys."""
    # params_flat = _flatten_dict(params)
    # expected_flat = _flatten_dict(expected)
    # missing_keys = expected_flat.keys() - params_flat.keys()
    # extra_keys = params_flat.keys() - expected_flat.keys()

    # Adds back empty dict explicitly, to support layers without weights.
    # Context: FLAX ignores empty dict during serialization.
    # empty_keys = set()
    # for k in missing_keys:
    #     if isinstance(expected_flat[k], dict) and not expected_flat[k]:
    #         params[k] = {}
    #         empty_keys.add(k)
    # missing_keys -= empty_keys

    # if empty_keys:
    # print("Inspect recovered empty keys:\n%s", empty_keys)
    # if missing_keys:
    #     print("Inspect missing keys:\n%s", missing_keys)
    # if extra_keys:
    #     print("Inspect extra keys:\n%s", extra_keys)
    #     raise ValueError(
    #         f"Missing params from checkpoint: {missing_keys}.\n"
    #         f"Extra params in checkpoint: {extra_keys}.\n"
    #         f"Restored params from checkpoint: {params_flat.keys()}.\n"
    #         f"Expected params from code: {expected_flat.keys()}."
    # )
    return params


def recover_tree(keys, values):
    """Recovers a tree as a nested dict from flat names and values.

    This function is useful to analyze checkpoints that are without need to access
    the exact source code of the experiment. In particular, it can be used to
    extract an reuse various subtrees of the scheckpoint, e.g. subtree of
    parameters.

    Args:
      keys: a list of keys, where '/' is used as separator between nodes.
      values: a list of leaf values.

    Returns:
      A nested tree-like dict.
    """
    tree = {}
    sub_trees = collections.defaultdict(list)
    for k, v in zip(keys, values):
        if "/" not in k:
            tree[k] = v
        else:
            k_left, k_right = k.split("/", 1)
            sub_trees[k_left].append((k_right, v))
    for k, kv_pairs in sub_trees.items():
        k_subtree, v_subtree = zip(*kv_pairs)
        tree[k] = recover_tree(k_subtree, v_subtree)
    return tree


def load(path):
    """Loads params from a checkpoint previously stored with `save()`."""
    with open(path, "rb") as f:
        ckpt_dict = np.load(f, allow_pickle=False)
        keys, values = zip(*list(ckpt_dict.items()))
    return recover_tree(keys, values)


def load_pretrained(*, pretrained_path, init_params, model_config, logger):
    """Loads/converts a pretrained checkpoint for fine tuning.

    Args:
      pretrained_path: File pointing to pretrained checkpoint.
      init_params: Parameters from model. Will be used for the head of the model
        and to verify that the model is compatible with the stored checkpoint.
      model_config: Configuration of the model. Will be used to configure the
        head and rescale the position embeddings.
      logger: Logger to use to output diagnostic messages.

    Returns:
      Parameters like `init_params`, but loaded with pretrained weights from
      `pretrained_path` and adapted accordingly.
    """

    restored_params = inspect_params(
        params=load(pretrained_path),
        expected=None,
        logger=logger,
        fail_if_extra=False,
        fail_if_missing=False,
    )

    # The following allows implementing fine-tuning head variants depending on the
    # value of `representation_size` in the fine-tuning job:
    # - `None` : drop the whole head and attach a nn.Linear.
    # - same number as in pre-training means : keep the head but reset the last
    #    layer (logits) for the new task.
    if model_config.representation_size is None:
        if "pre_logits" in restored_params:
            print("load_pretrained: drop-head variant")
            restored_params["pre_logits"] = {}
    # restored_params["head"]["kernel"] = init_params["head"]["kernel"]
    # restored_params["head"]["bias"] = init_params["head"]["bias"]

    if "posembed_input" in restored_params.get("Transformer", {}):
        pass
        # Rescale the grid of position embeddings. Param shape is (1,N,1024)
        # posemb = restored_params["Transformer"]["posembed_input"]["pos_embedding"]
        # posemb_new = init_params["Transformer"]["posembed_input"]["pos_embedding"]
        # if posemb.shape != posemb_new.shape:
        # pass
        # print(
        #     "load_pretrained: resized variant: %s to %s",
        #     posemb.shape,
        #     posemb_new.shape,
        # )
        # ntok_new = posemb_new.shape[1]

        # if model_config.classifier == "token":
        #     posemb_tok, posemb_grid = posemb[:, :1], posemb[0, 1:]
        #     ntok_new -= 1
        # else:
        #     posemb_tok, posemb_grid = posemb[:, :0], posemb[0]

        # gs_old = int(np.sqrt(len(posemb_grid)))
        # gs_new = int(np.sqrt(ntok_new))
        # print("load_pretrained: grid-size from %s to %s", gs_old, gs_new)
        # posemb_grid = posemb_grid.reshape(gs_old, gs_old, -1)

        # zoom = (gs_new / gs_old, gs_new / gs_old, 1)
        # posemb_grid = scipy.ndimage.zoom(posemb_grid, zoom, order=1)
        # posemb_grid = posemb_grid.reshape(1, gs_new * gs_new, -1)
        # posemb = jnp.array(np.concatenate([posemb_tok, posemb_grid], axis=1))
        # restored_params["Transformer"]["posembed_input"]["pos_embedding"] = posemb

    return restored_params


# my own code below:


def get_testing():
    """Returns a minimal configuration for testing."""
    config = ml_collections.ConfigDict()
    config.patches = ml_collections.ConfigDict({"size": (16, 16)})
    config.hidden_size = 1
    config.transformer = ml_collections.ConfigDict()
    config.transformer.mlp_dim = 1
    config.transformer.num_heads = 1
    config.transformer.num_layers = 1
    config.transformer.attention_dropout_rate = 0.0
    config.transformer.dropout_rate = 0.1
    config.classifier = "token"
    config.representation_size = None
    return config


def get_b16_config():
    """Returns the ViT-B/16 configuration."""
    config = ml_collections.ConfigDict()
    config.patches = ml_collections.ConfigDict({"size": (16, 16)})
    config.hidden_size = 768
    config.transformer = ml_collections.ConfigDict()
    config.transformer.mlp_dim = 3072
    config.transformer.num_heads = 12
    config.transformer.num_layers = 12
    config.transformer.attention_dropout_rate = 0.0
    config.transformer.dropout_rate = 0.1
    config.classifier = "token"
    config.representation_size = None
    return config


def get_b32_config():
    """Returns the ViT-B/32 configuration."""
    config = get_b16_config()
    config.patches.size = (32, 32)
    return config


def get_l16_config():
    """Returns the ViT-L/16 configuration."""
    config = ml_collections.ConfigDict()
    config.patches = ml_collections.ConfigDict({"size": (16, 16)})
    config.hidden_size = 1024
    config.transformer = ml_collections.ConfigDict()
    config.transformer.mlp_dim = 4096
    config.transformer.num_heads = 16
    config.transformer.num_layers = 24
    config.transformer.attention_dropout_rate = 0.0
    config.transformer.dropout_rate = 0.1
    config.classifier = "token"
    config.representation_size = None
    return config


def get_l32_config():
    """Returns the ViT-L/32 configuration."""
    config = get_l16_config()
    config.patches.size = (32, 32)
    return config


def get_h14_config():
    """Returns the ViT-H/14 configuration."""
    config = ml_collections.ConfigDict()
    config.patches = ml_collections.ConfigDict({"size": (14, 14)})
    config.hidden_size = 1280
    config.transformer = ml_collections.ConfigDict()
    config.transformer.mlp_dim = 5120
    config.transformer.num_heads = 16
    config.transformer.num_layers = 32
    config.transformer.attention_dropout_rate = 0.0
    config.transformer.dropout_rate = 0.1
    config.classifier = "token"
    config.representation_size = None
    return config


MODEL_SIZES = {
    "ViT-B_16": 86_567_656,
    "ViT-B_32": 88_224_232,
    "ViT-L_16": 304_326_632,
    "ViT-L_32": 306_535_400,
    "ViT-H_14": 632_045_800,
    "testing": 2985,
}


def vit_jax_to_pytorch(model_config, train_config, config, pathes_config):
    vit_variant = model_config.vit_variant
    vit_pretrained_dir = pathes_config.vit_pretrained_dir
    vitfp = os.path.join(vit_pretrained_dir, vit_variant + ".npz")
    vittorchfp = os.path.join(vit_pretrained_dir, "pytorch_model.bin")
    os.makedirs(vit_pretrained_dir, exist_ok=True)

    # rng = np.zeros(2)
    # VisualTransformer = KNOWN_MODELS[vit_variant].partial(num_classes=1000)
    # output, initial_params = VisualTransformer.init_by_shape(
    #     rng, [((2, 832, 832, 3), np.float32)]
    # )
    # init_flat = flatten_dict(initial_params)
    params = load_pretrained(
        pretrained_path=vitfp,
        init_params=None,
        model_config=get_b32_config(),
        logger=None,
    )
    vittorch = VisionTransformerPytorch(image_size=(384, 384), patch_size=(16, 16))
    vitdict = vittorch.state_dict()
    torch_from_jax = convert_jax_to_torch_weights(torch_dict=vitdict, jax_dict=params)
    vittorch.load_state_dict(torch_from_jax)
    torch.save(vittorch.state_dict(), vittorchfp)


if __name__ == "__main__":
    # from google.colab import drive

    # drive.mount("/gdrive")
    # root = "/gdrive/My Drive/vision_transformer_colab"
    # import os

    # if not os.path.isdir(root):
    #     os.mkdir(root)
    #     os.chdir(root)
    # print(f'\nChanged CWD to "{root}"')
    model_config = ModelConfig
    train_config = TrainConfig
    config = Config
    pathes_config = PathesConfig
    vit_jax_to_pytorch(model_config, train_config, config, pathes_config)



from setuptools import setup

from vltk.abc.extraction import VisnExtraction
from vltk.abc.visnadapter import VisnDataset
from vltk.abc.visnlangadatper import VisnLangDataset
from vltk.adapters import Adapters
from vltk.configs import DataConfig, LangConfig, VisionConfig
# from vltk.datasets import build
from vltk.dataset.builder import init_datasets
from vltk.features import Features
from vltk.processing import (LangProcessor, Processors, VisnLangProcessor,
                             VisnProcessor)
from vltk.vars import *


def build(config):
    return init_datasets(config)


"""
ALL BOXES ARE EXPECTED TO GO IN: (X, Y, W, H) FORMAT
"""


"""
:)
"""



from itertools import chain

import torch
import vltk.vars as vltk
from vltk.processing import VisnLangProcessor
from vltk.utils.adapters import truncate_and_pad_list


class Span(VisnLangProcessor):
    _keys = (vltk.tokenmap, vltk.span)

    def map_span(self, span, tokenmap, max_len):
        span = list(chain(*map(lambda x: [x[0]] * x[1], zip(span, tokenmap))))
        span = truncate_and_pad_list(span, max_len, 0)
        if self.config.add_visual_cls:
            span = [0] + span[1:]
        return span

    def forward(self, lang_entry, visn_entry, **kwargs):
        img_first = kwargs.get("img_first")
        max_len = self.config.lang.max_visual_seq_length
        tokenmap = visn_entry[vltk.tokenmap]
        span = lang_entry[vltk.span]
        if img_first:
            span = list(map(lambda s: self.map_span(s, tokenmap, max_len), span))
        else:
            span = self.map_span(span, tokenmap, max_len)
        lang_entry[vltk.span] = torch.tensor(span)
        return lang_entry, visn_entry



import json
from copy import deepcopy
from typing import Dict, List, Union

import torch
import vltk.vars as vltk
from torch.utils.data import DataLoader
from vltk.configs import DataConfig
from vltk.dataset.visndataset import VisionDataset
from vltk.dataset.visnlangdataset import VisionLanguageDataset


class BatchInfo:
    def __init__(self):
        self.uniq_keys = set()
        self.min_spanning_keys = set()
        self.sparse_keys = set()
        self.visn_keys = set()
        self.lang_keys = set()

    def update_entry_keys(self, entry):
        entry_keys = set(entry.keys())
        self.uniq_keys = self.uniq_keys.union(entry_keys)
        if not self.min_spanning_keys:
            self.min_spanning_keys = self.min_spanning_keys.union(entry_keys)
        else:
            self.min_spanning_keys = self.min_spanning_keys.intersection(entry_keys)
        self.sparse_keys = self.uniq_keys - self.min_spanning_keys

    def update_visn_lang_keys(self, lang_entry, visn_entry):
        self.visn_keys = self.visn_keys.union(set(visn_entry.keys()))
        self.lang_keys = self.lang_keys.union(set(lang_entry.keys()))

    def clear(self):
        self.uniq_keys.clear()
        self.min_spanning_keys.clear()
        self.sparse_keys.clear()

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        # rep = str(
        rep = json.dumps(
            {
                "uniq_keys": list(self.uniq_keys),
                "min_spanning_keys": list(self.min_spanning_keys),
                "sparse_keys": list(self.sparse_keys),
                "visn_keys": list(self.visn_keys),
                "lang_keys": list(self.lang_keys),
            },
            indent=4,
        )
        return rep

    @property
    def visn(self):
        return self.visn_keys

    @property
    def lang(self):
        return self.lang_keys


def collate(
    columns: List[Dict[str, torch.Tensor]],
    all_same_keys: bool = False,
    config: DataConfig = None,
    batch_info: BatchInfo = None,
) -> Dict[str, torch.Tensor]:

    if all_same_keys:
        batch = collate_homogeneous(columns, config)
    else:
        batch = collate_heterogenous(columns, config, batch_info)
    batch_info.clear()
    return batch


def collate_homogeneous(
    columns: List[Dict[str, torch.Tensor]],
    config: DataConfig,
):
    batch = {}
    # raise Exception
    for k in columns[0].keys():
        if k not in (vltk.imgid, vltk.qid):
            try:
                batch[k] = torch.stack([i.get(k) for i in columns if i is not None])
            except Exception:
                raise Exception(
                    "Each batch entry should all have the same keys in  `collate_homogeneous`",
                    f"Found sparse key: `{k}` that is only present in some entries. Try removing this key",
                )

        else:
            batch[k] = [i.get(k, None) for i in columns if i is not None]

    return batch


def collate_heterogenous(
    columns: List[Dict[str, torch.Tensor]], config: DataConfig, batch_info: BatchInfo
):
    batch = {}
    columns = sorted(columns, key=lambda x: len(x), reverse=True)
    if not config.collate_simple:
        for k in columns[0].keys():
            try:

                batch[k] = torch.stack([i.get(k) for i in columns if i is not None])
            except Exception:

                batch[k] = [i.get(k, None) for i in columns if i is not None]
    else:
        for k in batch_info.min_spanning_keys:
            try:
                batch[k] = torch.stack([i.get(k) for i in columns])
            except Exception:
                batch[k] = [i.get(k) for i in columns]

    return batch


def check_all_keys_same(config, visnlangdict=None, visndict=None, annodict=None):

    visn_keys_same = True
    visnlang_keys_same = True
    anno_keys_same = True
    tokenizer_in_visn_dataset = False
    visnlang_cols = set()
    visn_cols = set()
    anno_cols = set()
    if visnlangdict is not None:
        for dset in visnlangdict:
            for split in visnlangdict[dset]:
                adapter = visnlangdict[dset][split]
                if not visnlang_cols:
                    visnlang_cols = set(adapter.column_names)
                elif visnlang_cols != set(adapter.column_names):
                    visnlang_keys_same = False
                    visnlang_cols = visnlang_cols.union(adapter.column_names)
                else:
                    continue
    if visndict is not None:
        isadapter = True
        for dset in visndict:
            for split in visndict[dset]:
                dict_or_adapter = visndict[dset][split]
                if isinstance(dict_or_adapter, dict):
                    isadapter = False
                if not isadapter:
                    continue
                if not visn_cols:
                    visn_cols = set(dict_or_adapter.column_names)
                elif visn_cols != set(dict_or_adapter.column_names):
                    visn_keys_same = False
                    visn_cols = visn_cols.union(dict_or_adapter.column_names)
                else:
                    continue
            if not isadapter:
                continue
    if annodict is not None and not config.ignore_annotations:
        # TODO: check the case where segmentation things are different
        for adapter in annodict.values():
            if not anno_cols:
                # not so simple here, I must check that if there exists a difference, that the difference
                # only refers to a difference of segmentation datatypes
                try:
                    anno_cols = set(adapter.column_names)
                except TypeError:
                    anno_cols = set()

            try:
                adapter_cols = set(adapter.column_names)
            except Exception:
                adapter_cols = set()

            if vltk.text in adapter_cols:
                tokenizer_in_visn_dataset = True
            if anno_cols != adapter_cols:
                remaining_anno_cols = anno_cols - adapter_cols
                remaining_adapter_cols = adapter_cols - anno_cols
                remaining_cols = remaining_anno_cols.union(remaining_adapter_cols)
                if remaining_cols == set([vltk.polygons, vltk.RLE]):
                    continue
                anno_keys_same = False
                try:
                    anno_cols = visnlang_cols.union(adapter.column_names)
                except Exception:
                    pass
            replace_keys = anno_cols.intersection(visnlang_cols)
            if vltk.imgid in replace_keys:
                replace_keys.remove(vltk.imgid)
            anno_cols = anno_cols.union(set(map(lambda x: "v" + x, replace_keys)))
        if vltk.polygons in anno_cols or vltk.RLE in anno_cols:
            try:
                anno_cols.remove(vltk.polygons)
            except KeyError:
                pass
            try:
                anno_cols.remove(vltk.RLE)
            except KeyError:
                pass
            if not config.ignore_segmentation:
                anno_cols.add(vltk.segmentation)

    max_spanning_cols = visnlang_cols.union(visn_cols).union(anno_cols)
    print(
        f"Max spanning column names for each batch: {max_spanning_cols} (not including extra columns/features from processors)"
    )
    # here we will add all the feature names that could be added while we process each example
    all_same_keys = visnlang_keys_same and visn_keys_same and anno_keys_same
    if not all_same_keys and not config.collate_simple:
        print(
            "NOTICE: feature types differ between specificed datasets, so the batch dictionary will\
                                use a combination of the dataset names as keys which will point to their respective\
                                batch dictionaries."
        )
    return all_same_keys, max_spanning_cols, tokenizer_in_visn_dataset, replace_keys


class VisionLanguageLoader(DataLoader):
    def __init__(self, config, is_train=False, **kwargs):
        if not is_train:
            num_workers = 0
        else:
            num_workers = config.num_workers
        shuffle = config.shuffle if is_train else 0
        drop_last = config.drop_last
        pin_memory = config.pin_memory
        # init dataset

        visnlangdict = kwargs.get("visnlangdatasetadapterdict", None)
        annodict = kwargs.get("annotationdict", None)
        visndict = kwargs.get("visndatasetadapterdict", None)
        (
            all_same_keys,
            max_spanning_cols,
            tokenizer_in_visn_dataset,
            replace_keys,
        ) = check_all_keys_same(config, visnlangdict, visndict, annodict)
        kwargs["max_spanning_cols"] = max_spanning_cols
        batch_info = BatchInfo()
        self.batch_info = batch_info
        dataset = VisionLanguageDataset(
            config=config,
            is_train=is_train,
            batch_info=batch_info,
            tokenizer_in_visn_dataset=tokenizer_in_visn_dataset,
            **kwargs,
        )
        # init loader
        super().__init__(
            dataset=dataset,
            collate_fn=lambda x: collate(
                x,
                all_same_keys=all_same_keys,
                config=config,
                batch_info=batch_info,
            ),
            drop_last=drop_last,
            pin_memory=pin_memory,
            num_workers=num_workers,
            shuffle=shuffle,
            batch_size=dataset.batch_size,
        )

    def transpose_vl(self, batch, max_size=512):
        return self.dataset.transpose_vl(batch, max_size)


class VisionLoader(DataLoader):
    def __init__(self, config, is_train=False, **kwargs):
        if not is_train:
            num_workers = 0
        else:
            num_workers = config.num_workers
        shuffle = config.shuffle if is_train else 0
        drop_last = config.drop_last
        pin_memory = config.pin_memory
        # init dataset
        visnlangdict = kwargs.get("visnlangdatasetadapterdict", None)
        annodict = kwargs.get("annotationdict", None)
        visndict = kwargs.get("visndatasetadapterdict", None)
        batch_info = BatchInfo()

        (
            all_same_keys,
            max_spanning_cols,
            tokenizer_in_visn_dataset,
            replace_keys,
        ) = check_all_keys_same(config, visnlangdict, visndict, annodict)
        dataset = VisionDataset(
            config=config,
            is_train=is_train,
            batch_info=batch_info,
            tokenizer_in_visn_dataset=tokenizer_in_visn_dataset,
            **kwargs,
        )

        # init loader
        super().__init__(
            dataset=dataset,
            collate_fn=lambda x: collate(
                x,
                all_same_keys=all_same_keys,
                config=config,
                batch_info=batch_info,
            ),
            drop_last=drop_last,
            pin_memory=pin_memory,
            num_workers=num_workers,
            shuffle=shuffle,
            batch_size=dataset.batch_size,
        )



import os
from abc import abstractmethod
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
from typing import List

import datasets
import datasets as ds
import pyarrow
import vltk.vars as vltk
from datasets import ArrowWriter
from tqdm import tqdm
from vltk.abc.adapter import Adapter
from vltk.features import Features
from vltk.inspection import collect_args_to_func
from vltk.utils import base as utils
from vltk.utils.base import get_list_primitive


class VisnLangDataset(Adapter):
    _extensions = ["json", "jsonl"]
    _base_features = {
        vltk.imgid: Features.String(),
        vltk.text: Features.String(),
        # vltk.score: ds.Sequence(length=-1, feature=ds.Value("float32")),
    }
    _meta_names = ["answer_frequencies", "img_to_row_map"]
    _batch_size = 1028

    @staticmethod
    def adjust_imgid(img_id, dataset_name=None, split_name=None):
        """
        Sometimes the image IDS provided in Vision-Language datasets do not map to the
        actual image ID in the reffered vision dataset. If that is the case, implementing
        this optional function will adjust the image id in aims to appropriately mactch
        the corresponging dataset
        """
        return img_id

    @staticmethod
    def _check_features(schema):
        feature_dict = schema
        feature_dict[vltk.imgid] = VisnLangDataset._base_features[vltk.imgid]
        feature_dict[vltk.score] = VisnLangDataset._base_features[vltk.score]
        feature_dict[vltk.text] = VisnLangDataset._base_features[vltk.text]
        feature_dict[vltk.label] = VisnLangDataset._base_features[vltk.label]
        features = ds.Features(feature_dict)
        return features

    @staticmethod
    def _label_handler(label):
        single_label_score = [1]
        if isinstance(label, str):
            return [label], single_label_score
        if isinstance(label, dict):
            if len(label) == 0:
                return [""], single_label_score
            elif len(label) == 1:
                label = next(iter(label))
                assert isinstance(label, str), label
                return [label], single_label_score
            else:
                labels = []
                scores = []
                for lab, score in label.items():
                    assert isinstance(lab, str)
                    score = float(score)
                    labels.append(lab)
                    scores.append(score)
                return labels, scores

    @staticmethod
    def _locate_text_files(searchdir, textset_name, split):
        if isinstance(searchdir, list):
            searchdir = searchdir[0]
        searchdir = os.path.join(searchdir, textset_name)
        assert os.path.exists(searchdir)
        text_files = []
        suffixes = VisnLangDataset._extensions
        for datadir in [searchdir]:
            for suffix in suffixes:
                for path in Path(datadir).glob(
                    f"**/*.{suffix}",
                ):
                    path = str(path)
                    if textset_name in path.lower():
                        if split == "test" and "dev" in path:
                            continue
                        if split is None or split in path:
                            text_files.append(path)

        if not text_files:
            return None
        text_files = list(set(text_files))
        return text_files

    @staticmethod
    def _locate_text_set(datadir, textset_name, split):
        search_files = []
        for dd in [datadir]:
            search_files.append(os.path.join(dd, textset_name, f"{split}.arrow"))
        valid_search_files = list(filter(lambda x: os.path.isfile(x), search_files))
        assert any(valid_search_files), (
            f"attempting to load *.arrow datasets from the following pathes: '{search_files}'"
            f" but none of these are real files"
        )
        assert (
            len(valid_search_files) == 1
        ), "not sure which file to load: {valid_search_files}"
        valid_search_file = valid_search_files[0]
        return valid_search_file

    @classmethod
    def extract(
        cls,
        searchdir,
        config=None,
        splits=None,
        supervised=True,
        savedir=None,
        min_label_frequency=9,
        label_preprocessor="label_default",
        **kwargs,
    ):
        test_features = None
        if supervised:
            if min_label_frequency is None:
                assert config is not None
                min_label_frequency = config.min_label_frequency
            kwargs["min_label_frequency"] = min_label_frequency

        if splits is None:
            splits = vltk.SPLITALIASES
        else:
            if isinstance(splits, str):
                splits = [splits]

        if searchdir is None:
            searchdir = config.datadirs

        if savedir is None:
            savedir = os.path.join(searchdir, cls.__name__.lower())
        os.makedirs(savedir, exist_ok=True)

        print(f"searching for input files for splits: {splits}")
        file_split_dict = {}
        split_file_numbers = {}
        found_any_files = False
        for split in splits:

            text_files = cls._locate_text_files(
                searchdir=searchdir, textset_name=cls.__name__.lower(), split=split
            )
            if text_files is None:
                continue
            if hasattr(cls, "filters"):
                assert isinstance(
                    cls.filters, list
                ), f"filters must be in a list, not type {type(cls.filters)}"
                temp = []
                for i, t in enumerate(text_files):
                    passe_all = True
                    for f in cls.filters:
                        if f in t:
                            passe_all = False
                    if passe_all:
                        temp.append(t)

                text_files = temp
            if text_files:
                found_any_files = True
            file_split_dict[split] = text_files
            split_file_numbers[split] = len(text_files)

        if not found_any_files:
            print(
                "No files pattern matched with corresponding split, falling back to searching all json in top level directory"
            )
            text_files = cls._locate_text_files(
                searchdir=searchdir, textset_name=cls.__name__.lower(), split=None
            )
            if hasattr(cls, "filters"):
                assert isinstance(
                    cls.filters, list
                ), f"filters must be in a list, not type {type(cls.filters)}"
                temp = []
                for i, t in enumerate(text_files):
                    passe_all = True
                    for f in cls.filters:
                        if f in t:
                            passe_all = False
                    if passe_all:
                        temp.append(t)

                text_files = temp
            if not text_files:
                raise Exception("...could not locate any json files")
            file_split_dict["train"] = text_files

        split_dict = {}
        for split, text_files in file_split_dict.items():
            cur_row = 0
            imgid2rows = defaultdict(list)

            if not text_files:
                continue

            schema_dict = collect_args_to_func(cls.schema, kwargs=kwargs)
            features = ds.Features({**cls.schema(**schema_dict), **cls._base_features})
            meta_dict = VisnLangDataset._init_metadata(features)
            if not supervised:
                features.pop(vltk.score)
                features.pop(vltk.label)
            # setup arrow writer
            buffer = pyarrow.BufferOutputStream()
            stream = pyarrow.output_stream(buffer)
            if split == "test" or not supervised:
                if test_features is None:
                    test_features = deepcopy(features)
                    for f in deepcopy(set(test_features.keys())):
                        if f in meta_dict:
                            test_features.pop(f)
                writer = ArrowWriter(features=test_features, stream=stream)
            else:
                writer = ArrowWriter(features=features, stream=stream)
            # load data
            text_data = {}
            print(f"loading json files from: {text_files}")
            for t in tqdm(text_files):
                data = utils.try_load(t)

                text_data[t] = data

            # custom forward from user
            print("begin extraction")

            kwargs["datadir"] = searchdir
            forward_dict = collect_args_to_func(cls.forward, kwargs=kwargs)
            batch_entries = cls.forward(text_data, split, **forward_dict)

            # pre-checks
            print("writing rows to arrow dataset")
            for sub_batch_entries in utils.batcher(batch_entries, n=64):
                flat_entry = None
                for b in sub_batch_entries:
                    # apply adjust image id functio  here
                    vision_dataset_name_and_split = cls.data_info[split]
                    vdset_name = next(iter(vision_dataset_name_and_split.keys()))
                    vdset_split = next(iter(vision_dataset_name_and_split.values()))
                    b[vltk.imgid] = cls.adjust_imgid(
                        b[vltk.imgid], vdset_name, vdset_split
                    )
                    meta_dict = VisnLangDataset._update_metadata(meta_dict, b)

                    imgid2rows[b[vltk.imgid]].append(cur_row)
                    cur_row += 1
                    b = {k: [v] for k, v in b.items()}

                    if flat_entry is None:
                        flat_entry = b
                    else:
                        for k in flat_entry:
                            flat_entry[k].extend(b[k])

                if split == "test" or not supervised:

                    for f in deepcopy(set(flat_entry.keys())):
                        if f not in test_features:
                            flat_entry.pop(f)

                        if (
                            f in flat_entry
                            and get_list_primitive(flat_entry[f]) is None
                        ):
                            flat_entry.pop(f)
                            if f in test_features:
                                test_features.pop(f)

                    batch = test_features.encode_batch(flat_entry)
                    writer.write_batch(batch)
                else:
                    batch = features.encode_batch(flat_entry)
                    writer.write_batch(batch)

            # misc.
            savefile = os.path.join(savedir, f"{split}.arrow")
            meta_dict["img_to_row_map"] = imgid2rows
            meta_dict["split"] = split

            table, info, meta_dict = Adapter._save_dataset(
                buffer, writer, savefile, meta_dict, split
            )

            # return class
            arrow_dset = cls(
                arrow_table=table,
                split=split,
                info=info,
                meta_dict=meta_dict,
            )
            split_dict[split] = arrow_dset
        return split_dict

    def text_iter(self):
        for i in range(len(self)):
            row = self.get_row(i)
            if row:
                yield row

    def text_first(self):
        text_data = []
        for i in tqdm(range(len(self))):
            row = self.get_row(i)
            text_data.append(row)
        return text_data

    def get_frequency(self, label):
        return self.answer_frequencies[label]

    @property
    def split(self):
        try:
            return self._split
        except AttributeError:
            return None

    @property
    @abstractmethod
    def data_info(self) -> dict:
        raise Exception("Do not call from abstract class")

    @abstractmethod
    def forward(text_data: List[dict], split: str, **kwargs) -> List[dict]:
        pass

    @property
    def labels(self):
        return set(self.answer_frequencies.keys())

    @property
    def n_labels(self):
        return len(self.labels)

    @property
    def uniq_imgs(self):
        return set(self._img_to_row_map.keys())

    @property
    def answer_frequencies(self):
        return self._answer_frequencies



import itertools
import json
import logging as logger
import os
import pickle
import shutil
import sys
import tarfile
import zipfile
from abc import ABCMeta, abstractmethod
from collections import Counter, defaultdict
from pathlib import Path
from typing import List, Union

import datasets as ds
import pyarrow
import vltk.vars as vltk
import wget
from datasets import ArrowWriter, Dataset
from vltk.features import Features
from vltk.inspection import collect_args_to_func
from vltk.utils.base import (flatten_stringlist, get_arrow_primitive,
                             set_metadata)

IMGFILES = ("jpeg", "jpg", "png")
SUFFIXES = ("pdf", "json", "jsonl", "csv", "tsv")


class Adapter(ds.Dataset, metaclass=ABCMeta):

    filters: Union[List, None] = None
    """
    Adapter Property that stores a list of string that indicate which patterns
    should be avoided when the Adapter searches for relavant files.
    """
    urls: Union[List, str, None] = None
    """Adapter Property that stores a list of string or just a string which link to
    download urls.
    """
    _extensions = SUFFIXES
    _batch_size = 32
    _base_schema = {vltk.imgid: Features.Imgid()}
    _id_keys = {vltk.imgid, vltk.qid, vltk.text}
    _is_annotation = False
    _is_feature = False

    def __init__(
        self,
        arrow_table,
        meta_dict=None,
        split=None,
        info=None,
        fingerprint_off=False,
        **kwargs,
    ):
        if fingerprint_off:
            fingerprint = ""
        else:
            fingerprint = None

        super().__init__(
            arrow_table=arrow_table,
            split=split,
            info=info,
            fingerprint=fingerprint,
            **kwargs,
        )

        if meta_dict is None:
            meta_dict = {}
        for k, v in meta_dict.items():
            if isinstance(k, str):
                k_decoded = k
            else:
                k_decoded = k.decode()
            if k_decoded == "img_to_row_map" or k_decoded == "vocab":
                setattr(self, "_" + k_decoded, v)
            else:
                setattr(self, "meta_" + k_decoded, v)
        setattr(self, "_meta_dict", meta_dict)

    @classmethod
    def download(cls, datadir):
        name = cls.__name__.lower()
        assert (
            cls.urls is not None
        ), f"Cannot re/download {name} because the property: `urls`\
                is not defined. Please define this property as a string or a list of strings\
                in the adapter"
        urls = cls.urls
        if isinstance(urls, str):
            urls = [urls]

        def bar_progress(current, total, width=80):
            progress_message = "Downloading: %d%% [%d / %d] bytes" % (
                current / total * 100,
                current,
                total,
            )
            # Don't use print() as it will print in new line every time.
            sys.stdout.write("\r" + progress_message)
            sys.stdout.flush()

        path = Path(os.path.join(datadir, name))
        temppath = Path(os.path.join(datadir, f"temp_{name}"))
        try:
            if path.exists():
                temppath.mkdir()
                for f in path.glob("**/*"):
                    newpath = temppath.joinpath(f.name)
                    f.rename(newpath)

                if not path.exists():
                    path.mkdir()
            else:
                os.makedirs(str(path), exist_ok=True)
            for link in urls:
                print(f"Downloading {link}")
                filename = wget.download(link, bar=bar_progress, out=str(path))
                print()

                if ".tar.gz" in filename:
                    with tarfile.open(filename) as tar_ref:
                        print(f"Extracting {filename}")
                        tar_ref.extractall(path)

                elif ".zip" in filename:
                    with zipfile.ZipFile(filename, "r") as zip_ref:
                        print(f"Extracting {filename}")
                        zip_ref.extractall(path)
                else:
                    raise Exception(f"{filename} is not a zip file or tarball")
                    # name = Path(filename).name
                    # shutil.move(filename, os.path.join(path, name))
                    # os.remove(filename)
            shutil.rmtree(str(temppath), ignore_errors=True)
            shutil.rmtree(os.path.join(str(path), "__MACOSX"), ignore_errors=True)
        except Exception as e:
            print("WARNING: therew was an exception: {e}")
            # now remove temppath
            if temppath.exists():
                ###
                for f in temppath.glob("**/*"):
                    newpath = path.joinpath(f.name)
                    f.rename(newpath)
                ###
                if temppath.exists():
                    try:
                        os.rmdir(str(temppath))
                    except Exception:
                        raise Exception(
                            f"{temppath} was made in a failed download attempt. Attempting to remove path,\
                                 but, {temppath} is not empty."
                        )
            raise Exception(
                f"Download Failed: {e}. Moving {str(temppath)} back to {str(path)}"
            )

    @property
    def meta_dict(self):
        return self._meta_dict

    def get_metadata_counters(self):
        if not self._meta_dict:
            return {}

        try:
            schema_dict = collect_args_to_func(type(self).schema, kwargs={})
            feature_dict = {**type(self).schema(**schema_dict), **self._base_features}
        except ValueError:
            feature_dict = {**type(self).schema(), **self._base_features}
        counter_keys = tuple(self._init_metadata(feature_dict).keys())
        counters = {}
        for key in counter_keys:
            if not isinstance(next(iter(self._meta_dict)), str):
                key_encoded = key.encode()
            else:
                key_encoded = key
            if key_encoded in self._meta_dict:
                counters[key] = self._meta_dict[key_encoded]
        return counters

    def has(self, img_id):
        return img_id in self.img_to_row_map

    def get(self, img_id, return_dataset=False):
        if not return_dataset:
            return self[self.img_to_row_map[img_id]]
        else:
            idxs = self.img_to_row_map[img_id]
            assert isinstance(idxs, list)
            return self.select(idxs)

    def get_idx(self, img_id):
        return self.img_to_row_map[img_id]

    def shuffle(self):
        raise NotImplementedError

    def imgid_filter(self, imgids, is_visnlang=True):
        remaining = set(self.imgids).intersection(imgids)

        if is_visnlang:
            idx_groups = dict(
                map(lambda imgid: (imgid, self.get_idx(imgid)), remaining)
            )
            new_map = defaultdict(list)
            idx = 0
            idx_set = []
            for imgid, idxs in idx_groups.items():
                if isinstance(idxs, int):
                    idxs = [idxs]
                    # raise Exception(idx_set, idxs, self)
                idx_set.extend(idxs)
                new_map[imgid] = list(map(lambda x: x[0] + idx, enumerate(idxs)))
                idx += len(idxs)
        else:
            idx_set = list((map(lambda idx: self.get_idx(idx), remaining)))
        filtered_self = self.select(idx_set)
        setattr(filtered_self, "img_to_row_map", self.img_to_row_map)
        setattr(filtered_self, "get_metadata_counters", self.get_metadata_counters)
        setattr(filtered_self, "get", self.get)
        if is_visnlang:
            try:
                setattr(filtered_self, "data_info", self.data_info)
            except Exception:
                pass
            setattr(filtered_self, "_img_to_row_map", new_map)
            try:
                setattr(filtered_self, "imgids", remaining)
            except Exception:
                pass

            # setattr(filtered_self, "_img_to_row_map", new_map)
        else:
            setattr(filtered_self, "_img_to_row_map", dict(zip(remaining, idx_set)))
            setattr(filtered_self, "check_imgid_alignment", self.check_imgid_alignment)
            setattr(filtered_self, "imgids", remaining)

        return filtered_self

    @property
    def img_to_row_map(self):
        return self._img_to_row_map

    @property
    def name(self):
        return type(self).__name__.lower()

    @property
    def imgids(self):
        return tuple(self._img_to_row_map.keys())

    @property
    def n_imgs(self):
        return len(self.imgids)

    @staticmethod
    def _custom_finalize(writer, close_stream=True):
        if writer.pa_writer is None:
            if writer._schema is not None:
                writer._build_writer(writer._schema)
            else:
                raise ValueError(
                    "Please pass `features` or at least one example when writing data"
                )
        writer.pa_writer.close()
        if close_stream:
            writer.stream.close()
        logger.info(
            "Done writing %s %s in %s bytes %s.",
            writer._num_examples,
            writer.unit,
            writer._num_bytes,
            writer._path if writer._path else "",
        )
        return writer._num_examples, writer._num_bytes

    @staticmethod
    def _get_valid_search_pathes(searchdir, name=None, splits=None, annodir=None):
        if splits is None:
            splits = vltk.SPLITALIASES
        elif isinstance(splits, str):
            splits = [splits]
        assert os.path.isdir(
            searchdir
        ), f"The specificed datadir, {searchdir}, does not exist"
        if name is not None:
            searchdir = os.path.join(searchdir, name)
            assert os.path.isdir(searchdir), f"{searchdir} is not a dir"
        if annodir is not None:
            return searchdir, None
        final_paths = []
        valid_splits = []
        for splt in splits:
            path = os.path.join(searchdir, splt)
            if not os.path.isdir(path):
                continue
            final_paths.append(path)
            valid_splits.append(splt)

        assert final_paths, (searchdir, name, splits, annodir)
        return final_paths, valid_splits

    @staticmethod
    def _make_save_path(searchdir, dataset_name, extractor_name):
        if dataset_name is not None:
            savepath = os.path.join(searchdir, dataset_name, extractor_name)
        else:
            savepath = os.path.join(searchdir, extractor_name)
        print(f"will write to {savepath}")
        os.makedirs(savepath, exist_ok=True)
        return savepath

    @staticmethod
    def _iter_files(searchdirs, valid_splits=None, iter_imgs=False):
        text_files = []
        if isinstance(searchdirs, str):
            searchdirs = [searchdirs]
        for datadir in searchdirs:
            iterfilter = IMGFILES if iter_imgs else SUFFIXES
            for suffix in iterfilter:
                for path in Path(datadir).glob(
                    f"**/*.{suffix}",
                ):
                    path = str(path)
                    if valid_splits is not None:
                        split_in = False
                        for split in valid_splits:
                            if split in path:
                                split_in = True
                        if split_in:
                            text_files.append(path)
                    else:
                        text_files.append(path)
                    # if textset_name in path.lower():
                    #     if split == "test" and "dev" in path:
                    #         continue
                    #     if split is None or split in path:
                    #         text_files.append(path)

        if not text_files:
            return None
        text_files = list(set(text_files))
        return text_files

    @staticmethod
    def _build_schema(features_func, **kwargs):
        feat_args = collect_args_to_func(features_func, kwargs)
        features = features_func(**feat_args)
        default = Adapter._base_schema
        features = {**default, **features}
        return features

    @staticmethod
    def _save_dataset(buffer, writer, savefile, meta_dict, split=None):
        dset = Dataset.from_buffer(buffer.getvalue(), split=ds.Split(split))
        try:
            writer.finalize(close_stream=False)
        except Exception:
            pass
        # misc.
        dset = pickle.loads(pickle.dumps(dset))
        # add extra metadata
        table = set_metadata(
            dset._data, tbl_meta=meta_dict if meta_dict is not None else {}
        )
        # define new writer
        writer = ArrowWriter(path=savefile, schema=table.schema, with_metadata=False)
        # savedir new table
        writer.write_table(table)
        e, b = Adapter._custom_finalize(writer, close_stream=True)
        print(f"Success! You wrote {e} entry(s) and {b >> 20} mb")
        print(f"Located: {savefile}")
        return (table, dset.info, meta_dict)

    @staticmethod
    def _load_one_arrow(filestem, meta_names, config=None, name=None):
        if ".arrow" not in filestem:
            path = os.path.join(filestem, ".arrow")
        else:
            path = filestem
        if not os.path.isfile(path):
            path = path.replace("/annotations/", "/")
        assert os.path.isfile(path), f"{path} does not exist"
        # this is where we do the re-extract
        # if not os.path.isfile(path) or (
        #     config is not None and config.reextract and name is not None
        # ):
        #     datadir = config.datadir
        mmap = pyarrow.memory_map(path)
        f = pyarrow.ipc.open_stream(mmap)
        pa_table = f.read_all()
        meta_dict = {}
        for n in pa_table.schema.metadata.keys():
            if n.decode() == "huggingface":
                continue

            data_dump = pa_table.schema.metadata[n]
            try:
                data = json.loads(data_dump)
            except Exception:
                data = data_dump
            meta_dict[n] = data
        return (pa_table, meta_dict, path)

    @staticmethod
    def _load_many_arrows(stem, meta_names, config=None):
        split_list = []
        for split in vltk.SPLITALIASES:
            temppath = os.path.join(stem, f"{split}.arrow")
            if not os.path.isfile(temppath):
                continue
            pa_table, meta_dict, path = Adapter._load_one_arrow(
                temppath, meta_names, config=config
            )
            split_list.append((pa_table, meta_dict, split))
        return split_list

    @classmethod
    def load(cls, path, split=None, dataset_name=None, config=None):
        meta_names = cls._meta_names
        if ".arrow" in path:
            (pa_table, meta_dict, path) = Adapter._load_one_arrow(
                path, meta_names, config=config
            )
            return cls(arrow_table=pa_table, split=split, meta_dict=meta_dict)
        # to return visual features
        if dataset_name is not None:
            path = os.path.join(path, dataset_name)
        path = os.path.join(path, cls.__name__.lower())
        if cls._is_annotation:

            temppath = os.path.join(path, f"{vltk.ANNOTATION_DIR}.arrow")
            if not os.path.exists(temppath):
                temppath = os.path.join(
                    path, f"{vltk.ANNOTATION_DIR}/{vltk.ANNOTATION_DIR}.arrow"
                )
            path = temppath
            (pa_table, meta_dict, path) = Adapter._load_one_arrow(
                path, meta_names, config=config
            )
            return cls(arrow_table=pa_table, split=split, meta_dict=meta_dict)
        elif split is not None:
            path = os.path.join(path, f"{split}.arrow")
            (pa_table, meta_dict, path) = Adapter._load_one_arrow(
                path, meta_names, config=config
            )
            return cls(arrow_table=pa_table, split=split, meta_dict=meta_dict)
        else:
            arrow_dict = {}
            split_list = Adapter._load_many_arrows(path, meta_names, config=config)
            for sl in split_list:
                (pa_table, meta_dict, split) = sl
                arrow_dict[split] = cls(
                    arrow_table=pa_table, split=split, meta_dict=meta_dict
                )
            return arrow_dict

    @staticmethod
    def _init_metadata(schema):
        metadata_dict = {}
        for k, v in schema.items():
            if k not in Adapter._id_keys and get_arrow_primitive(v) == "string":
                metadata_dict[k] = Counter()
        return metadata_dict

    @staticmethod
    def _update_metadata(meta_dict, batch_dict):
        for k, v in meta_dict.items():
            if k in batch_dict:
                meta_dict[k].update(flatten_stringlist(batch_dict[k]))
        return meta_dict

    @abstractmethod
    def forward(*args, **kwargs):
        raise Exception("child forward method is not being called")

    @abstractmethod
    def schema(*args, **kwargs):
        return dict

    @property
    @abstractmethod
    def _meta_names():
        return List
